#!/usr/bin/perl
# $Id: psheduler.pl,v 1.4 1999/11/20 18:34:13 vasim Exp $
use Sys::Syslog qw(:DEFAULT setlogsock);

# Configuration variables
$tnpppagecommand = "/usr/local/bin/tnpppage";
$tnppaddr = "0002";
$tnppsocket = "/var/run/rtnppd.queue.out";
$spooldirectory = "/var/spool/psheduler";
$shedulerfunc = "0x4f";


setlogsock('unix');
openlog("psheduler", "ndelay|pid", "user");
if (!opendir(SPOOL, $spooldirectory)) {
  syslog('warning', 'Cannot open spool directory "%s"', $spooldirectory);
  closelog();
  exit(1);
}

$checkname = "$spooldirectory/last.check";
if (-f "$checkname") {
  ($xdev, $xino, $xmode, $xnlink, $xuid, $xgid, $xrdev, $xsize, $xatime,
   $xmtime, $xctime, $xblksize, $xblocks) = stat($checkname);
  ($xsec, $lastcheckmin, $lastcheckhour, $xmday, $xmon, $xyear, $xwday,
   $xyday, $xisdst) = localtime($xmtime);
} else {
  $lastcheckhour = -1;
  $lastcheckmin = -1;
}

@files = readdir(SPOOL);
($sc, $curmin, $curhour, $curday, $curmonth, $curyear, $curwday, $yd, $ds) =
  localtime();
$curmonth++;
$curyear = $curyear + 1980;

syslog('info', 'Checking spool directory');
foreach $filename (@files) {
  if ($filename =~ /^([^.]+)\.([A-Za-z0-9]+)\.(.+)$/) {
    syslog('debug', 'Processing file "%s"', $filename);
    $prefix = $1;
    $pager = $2;
    $name = $3;
    @tms = split(/;/, $name);
    $ntms = @tms;
    $resname = "";
    $changed = 0;
    
    $checktime = 0;
    for $i (0 .. ($ntms - 1)) {
      # Check each date/time of delivery
      $checkdate = 0;
      $checkwday = 0;
      $invalid = 0;
      
      $rec = @tms[$i];

      if ($rec =~ /^x(.*)$/) {
	$nodelete = 1;
	$rec = $1;
      } else {
	$nodelete = 0;
      }
      
      if ($rec =~ /^([^ 0-9]*) ?([0-9]+):([0-9]+)$/) {
	$checktime = 1;
	$hourdelivery = $2;
	$mindelivery = $3;
	if (($hourdelivery > 23) || ($mindelivery > 59)) {
	  $invalid = 1;
	}
	$rec = $1;
      }
      
      if ($rec =~ /^([0-9]+)\/([0-9]+)\/([0-9]+)$/) {
	$checkdate = 1;
	$daydelivery = $2;
	$monthdelivery = $1;
	if ($3 < 100) {
	  if ($3 > 98) {
	    $yeardelivery = $3 + 1900;
	  } else {
	    $yeardelivery = $3 + 2000;
	  }
	} else {
	  $yeardelivery = $3;
	}
	if (($yeardelivery < 1999) || ($monthdelivery > 12)
	    || ($monthdelivery < 1) || ($daydelivery > 31)
	    || ($daydelivery < 1)) {
	  $invalid = 1;
	}
      } else {
	if ($rec ne "") {
	  if ($rec =~ /^mo$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 1;
	  } elsif ($rec =~ /^tu$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 2;
	  } elsif ($rec =~ /^we$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 3;
	  } elsif ($rec =~ /^th$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 4;
	  } elsif ($rec =~ /^fr$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 5;
	  } elsif ($rec =~ /^sa$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 6;
	  } elsif ($rec =~ /^su$/i) {
	    $checkwday = 1;
	    $wdaydelivery = 0;
	  } else {
	    $invalid = 1;
	  }
	}
      }
      
      if ($invalid == 0) {
	$delivery = -1;
	if ($checkdate) {
	  # Check date
	  $delivery = 0;
	  if ($curyear > $yeardelivery) {
	    $delivery = 1;
	  } elsif ($curyear == $yeardelivery) {
	    if ($curmonth > $monthdelivery) {
	      $delivery = 1;
	    } elsif ($curmonth == $monthdelivery) {
	      if ($curday >= $daydelivery) {
		$delivery = 1;
	      }
	    }
	  }
	}
	if ($checkwday) {
	  # Check day of week
	  $delivery = 0;
	  if ($curwday == $wdaydelivery) {
	    $delivery = 1;
	  }
	}
	if ($checktime && (($delivery == -1) || ($delivery == 1))) {
	  # Check time
	  $delivery = 0;
	  if ($curhour > $hourdelivery) {
	    if (($hourdelivery > $lastcheckhour)
		|| (($hourdelivery == $lastcheckhour)
		    && ($mindelivery >= $lastcheckmin))) {
	      $delivery = 1;
	    }
	  } elsif ($curhour == $hourdelivery) {
	    if ($curmin >= $mindelivery) {
	      if (($hourdelivery > $lastcheckhour)
		  || (($hourdelivery == $lastcheckhour)
		      && ($mindelivery >= $lastcheckmin))) {
		$delivery = 1;
	      }
	    }
	  }
	}
	if ($delivery == 1) {
	  # Deliver page
	  $pagecommand = "$tnpppagecommand -t $tnppaddr -s $tnppsocket -c $shedulerfunc";
	  $res = system("$pagecommand -i $pager -f '$spooldirectory/$filename'");
	  $exitcode = $res / 256;
	  if ($exitcode != 0) {
	    syslog('warning', 'Exit code from tnpppage = %d', $exitcode);
	    $rec = @tms[$i];
	    $resname = "$resname$rec;";
	  } else {
	    if ($nodelete) {
	      $rec = @tms[$i];
	      $resname = "$resname$rec;";
	    } else {
	      $changed = 1;
	    }
	  }
	} else {
	  $rec = @tms[$i];
	  $resname = "$resname$rec;";
	}
      } else {
	# Invalid date/time definition
	syslog('warning', 'Error in time/date definition: "%s"', $rec);
      }
    }
    if ($changed) {
      if ("$resname" eq "") {
	syslog('info', 'Removing page "%s" for "%s"', $filename, $pager);
	unlink("$spooldirectory/$filename");
      } else {
	$newname = "$prefix.$pager.$resname";
	syslog('info', 'Renaming page "%s" to "%s"', $filename, $newname);
	rename("$spooldirectory/$filename", "$spooldirectory/$newname");
      }
    }
  }
}
syslog('info', 'Spool directory checked. Exiting');
open(XOUT, ">$checkname");
print XOUT "Don't delete this file, please!\n";
close(XOUT);
closelog();
exit(0);
