/* config file format & parse */
/* $Id: readcfg.h,v 1.4 1999/02/15 16:39:13 vasim Exp $ */

typedef enum { BOOL_NOP, BOOL_PAR, STRING, STRING_NC, STRING_WQ,
	       DEC_DIGITS, HEX_DIGITS, IGNORE } param_type;

typedef struct conf_rec {
  char           *param;	/* name of parameter (regular expression) */
  param_type      type;         /* type of parameter */
  void           *addr;		/* address of for parameter variable
				* (for STRING - address of string's pointer */
  int             len;		/* length for string parameters. If == 0,
				* malloc need for value */
  void           *def;		/* default value (for BOOL_NOP, BOOL_PAR,
				* DEC_DIGITS and HEX_DIGITS - int value.
				* for STRING - pointer on default string */
  int             flag_default; /* if 0 - default initialization not need */
  int (*value_func)(void *addr, int len, int flag, void *value);
}               conf_rec;

/* Maximum number of parameters */
#define MAX_CONF_PARAM_NUM 64

/* error text */
extern char *config_error;

/* Read and parse config file. If return 0 - was error in config file */
/* Do not use it in SUID/SGID programs! */
int read_config(char *filename, conf_rec params[], int number, char *section);

/* standart initialization functions */
/* flag - 0 for default initialization */
int std_STRING(void *addr, int len, int flag, void *value);
int std_DIGITS(void *addr, int len, int flag, void *value);
int std_BOOL(void *addr, int len, int flag, void *value);

/* Regexps for parse config file */
/* see man regexp(3) */
/* space(s) between parameter's name and value */
#define E_SPACE "[ \\\t=:]+"
/* space(s) before parameter's name */
#define B_SPACE "^[ \\\t]*"
/* value 'OFF' for BOOL_PAR */
#define F_NO "((no)|(off)|(sux+)|(0))"
/* value 'ON' for BOOL_PAR */
#define F_YES "((yes)|(on)|(rulez+)|(1))"
/* value for DEC_DIGITS */
#define F_DEC "[-+]?[0-9]+"
/* value for HEX_DIGITS */
#define F_HEX "((0x)|(x))?[0-9a-f]+"
/* value for STRING (without comment checking, space(s) allowed) */
#define F_STRING_NC "[!-~][ -~]*"
/* value for STRING (with comment checking, no space(s)) - recommended */
#define F_STRING "[!-\"$-:<-~]+"
/* value for STRING (in quotas. can have NULL value, if empty) */
#define F_STRING_WQ "['\"][^'\"]*['\"]"
/* comment line */
#define F_COMMENT "^[ \\\t]*(([;#])|([\\\r\\\n]*$))"

#define PARAM(x)  B_SPACE x E_SPACE
#define PARAM_NOP(x) B_SPACE x "[ \\\t]*"
