/* rtapd by Vasim V. */
/* $Id: rtapd.c,v 1.45 2001/12/02 06:58:45 vasim Exp $ */

#include <sys/types.h>
#include "queue.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <termios.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <termios.h>
#include <limits.h>
#include "crc.h"
#include "readcfg.h"
#include "rtapd.h"

#ifdef USE_MSQL
#include "msql.h"
#else
#ifdef USE_POSTGRESQL
#include "libpq-fe.h"
#else
#ifdef USE_MYSQL
#include "mysql.h"
#endif
#endif
#endif

#ifndef B7200
/* Heh... */
#define B7200 0
#endif

#ifndef TIOCFLUSH
#define TIOCFLUSH TCFLSH
#endif

#define L_FATAL LOG_CRIT
#define L_ERROR LOG_ERR
#define L_WARNING LOG_WARNING
#define L_INFO LOG_INFO
#define L_DEBUG LOG_DEBUG


STAILQ_HEAD(devhead, tap_device) head;

/* SQL-table format: */
/* XFROM int (for whom? 0, if for all) */
/* XWHERE char(10) */
/* XTO text (destination. One route can have multiply destinations) */

/* Devices names (in SQL-database): */
/* TNPP:<addr> (to rtnppd) */
/* PAGER:<addr> (forces query to database) */
/* TAP:DIAL<addr>:<phone> (call to <phone> over device(s) DIAL<addr>-
 * <addr> can contain simpy pattern '*') */
/* TAP:SERIAL<addr> (to serial TAP-device) */
/* NONE (drop packet) */

/* Packets to busy devices */
typedef struct wo_packet {
  tap_packet *pack;
  char device[128];
  STAILQ_ENTRY(wo_packet) packp;
} wo_packet;

STAILQ_HEAD(wpackhead, wo_packet) wpacks;
int nwpacks;

/* Routing table from file */
typedef struct fixed_route {
  int from;
  char to[22];
  char destination[256];
  STAILQ_ENTRY(fixed_route) next;
} fixed_route;

STAILQ_HEAD(routehead, fixed_route) routes;

int flag_fatal_error;
int flag_reread;
int loglevel;
int log_stderr;
int log_all_data;
int use_file_table;
int nodaemon;

char *lock_dir;         /* UUCP-locking directory */
char *file_dev_conf;    /* File with devices config */
char *file_main_conf;   /* Main config file */
char *file_route_table; /* Routing table file */
char *file_pid;         /* Pid-file name */
char *file_bad;         /* Logfile for unsuccessful pages */
char *section;          /* Section of config files */
char *command_filter;   /* External command for 'command block' */

tap_device *defaultroute;

int my_tnpp_addr;

int maxpacks;           /* Max. limit of packets in one device's queue */
int max_wopacks;        /* Max. limit of delayed packets */

char page_func = '@';
int sheduler_func;

char *sql_serv;
char *sql_base;
char *sql_port;
char *sql_user;
char *sql_passwd;
char *u_sock_name;
char *u_sock_name_out;
struct sockaddr_un u_sun;
int allowsend;              /* Allow send messages to all local users */

int use_sql;
#ifdef USE_MSQL
int sock_sql;               /* Socket to SQL-server */
#else
#ifdef USE_POSTGRESQL
PGconn *conn_sql;
#else
#ifdef USE_MYSQL
MYSQL *conn_sql = NULL;
#endif
#endif
#endif

/* External queue's sockets */
int tnpp_in;
int tnpp_out;

void recv_data(tap_device *dev, int len);  /* Process new data */
void send_data(tap_device *dev, int len);  /* Prepare to send new data */
void wlog(int prio, char *fmt, ...);
void log_data(char *type, unsigned char *data, int len);
tap_packet *create_pack();                 /* Create new empty packet */
void queue_ack(tap_device *dev, unsigned char *s);  /* Create message */
int checksum(unsigned char *s, int len);   /* Calculate TAP checksum */
void dequeue_head(tap_device *dev);        /* Remove first packet in queue */
void dequeue_message(tap_device *dev);     /* Remove all packets for first
					    * message in queue */
void log_msg(unsigned char *msg, int len, char *reason);
void set_port_parity(int fd, int bits, int parity);
                                           /* Set parity */
void set_serial(tap_device *dev);          /* Set serial parameters */
void reconnect(tap_device *dev);           /* Re-open serial port */
void toggle_dtr(int fd);
void check_wo_packs(tap_device *dev);      /* Check helded packets */
int uucp_lock(char *device);
void uucp_unlock(char *device);
int process_tap_packet(unsigned char *data, int len);
char *get_dest(char *dest);                /* Search route for dest */
char *run_query(char *query);              /* Make request to SQL server */
tap_packet *get_next_block(unsigned char *data, int *offset, int len);
                                           /* Create new block */
void conv_cr(tap_packet *pack, int type);  /* Map CR to CR+LF or LF+CR */
tap_device *check_avail_dev(char *dest, int *max_r_len);
                                   /* Search free device for dest */
int route_pack(char *dest, unsigned char *data, int len, char p_func);
                                           /* Route packet to dest */
void check_name(char *name);               /* Check filename for 'bad' chars
					    * Allow only 'A-Za-z0-9' */
int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		     int len, int tnpp_to, int tnpp_from);
int recv_tnpp_packet(int fd, struct sockaddr_un *sun);
void check_call_limit(tap_device *dev);     /* Check call's limit */
void block_concurrent(tap_device *dev);
void unblock_concurrent(tap_device *dev);
int to_sql_connect ();
void to_sql_disconnect();

#define CAPTYPES_NUMBER 30
struct {
  char *pgname;
  int pgtype;
  int caplen;
  int usedlen;
} captypes [CAPTYPES_NUMBER] = {
  {"POCSAG-512", 0x50, 8, 8},
  {"POCSAG-1200", 0x70, 8, 8},
  {"POCSAG-2400", 0x51, 8, 8},
  {"COLAY", 0x47, 8, 8},
  {"COLAY-II-A", 0x54, 8, 8},
  {"FLEX", 0x46, 13, 9},
  {"FLEX-AUGM", 0x66, 12, 10},
  {"ERMES", 0x65, 12, 8},
  {"ERMES-AUGM", 0x72, 19, 7},
  {"NEC-D2", 0x6e, 8, 8},
  {"NEC-D3", 0x4e, 8, 8},
  {"APOC-1200", 0x61, 10, 8},
  {"APOC-2400", 0x41, 10, 8},
  {"HSC", 0x48, 8, 8},
  {"5-6-TONE", 0x35, 8, 8},
  {"2-TONE", 0x32, 8, 8},
  {"RDS", 0x52, 8, 8},
  {"MBS", 0x4d, 8, 8},
  {"MTONE-IV", 0x34, 8, 8},
  {"MTONE-V", 0x56, 8, 8},
  {"MTONE-VI", 0x36, 8, 8},
  {"SPANTEL", 0x53, 8, 8},
  {"DTMF", 0x44, 8, 8},
  {"DTMF-NET", 0x64, 8, 8},
  {"ECHO", 0x45, 8, 8},
  {"NEWSPAGER-512", 0x5c, 8, 8},
  {"NEWSPAGER-1200", 0x7c, 8, 8},
  {"NEWSPAGER-2400", 0x2f, 8, 8},
  {"SA206", 0x4c, 8, 8},
  {"RESERVED", 0x2e, 8, 8}
};

#define CAPCLASS_NUMBER 15
struct {
  char *clname;
  int cltype;
} capclass[CAPCLASS_NUMBER] = {
  {"BEEP", 0x42},
  {"NUMERIC", 0x4e},
  {"SHIFTED", 0x6e},
  {"ALPHANUMERIC", 0x41},
  {"VOICE", 0x56},
  {"DATA", 0x44},
  {"CLASS-1", 0x31},
  {"CLASS-2", 0x32},
  {"CLASS-3", 0x33},
  {"CLASS-4", 0x34},
  {"CLASS-5", 0x35},
  {"CLASS-6", 0x36},
  {"CLASS-7", 0x37},
  {"CLASS-8", 0x38},
  {"CLASS-9", 0x39}
};

/* Calculate checksum */
int checksum(unsigned char *s, int len) {
  int i;
  int sum;

  sum = 0;
  for (i = 0; i < len; i++)
    sum += s[i];
  return (sum & 0x0fff);
} /* checksum */

/* Map CR to CR+LF or LF+CR in data packet */
void conv_cr(tap_packet *pack, int type) {
  unsigned char *p;

  pack->data[pack->size] = '\0';
  p = strchr(pack->data, CR);
  while (p) {
    if ((p != (pack->data + pack->size - 1))
	&& (p < (pack->data + sizeof(pack->data) - 3)))
      memmove(p + 1, p, pack->size - (p - pack->data));
    if (type) 
      *(p + 1) = LF;
    else {
      *p = LF;
      *(p + 1) = CR;
    }
    p = p + 2;
    pack->size++;
    if (p >= (pack->data + pack->size))
      break;
    p = strchr(p, CR);
  }
} /* conv_cr */

/* Get next 250 chars from data and create TAP-block */
tap_packet *get_next_block(unsigned char *data, int *offset, int len) {
  unsigned char *p;
  int size;
  tap_packet *pack;
  int csum;

  p = data + *offset;
  size = len - *offset;
  if (size > 250)
    size = 250;
  *offset += size;
  pack = create_pack();
  if (!pack)
    return NULL;
  pack->size = size + 6;
  pack->offset = 0;
  pack->status = P_DATA;
  pack->data[0] = STX;
  memcpy(pack->data + 1, p, size);
  if (*offset >= len) {
    if (pack->data[size] != CR) {
      pack->data[size + 1] = CR;
      size++;
      pack->size++;
    }
    pack->data[size + 1] = ETX;
  } else
    if (pack->data[size] == CR)
      pack->data[size + 1] = ETB;
    else
      pack->data[size + 1] = US;
  csum = checksum(pack->data, size + 2);
  pack->data[size + 2] = ((csum >> 8) & 0x0f) + 48;
  pack->data[size + 3] = ((csum >> 4) & 0x0f) + 48;
  pack->data[size + 4] = (csum & 0x0f) + 48;
  pack->data[size + 5] = CR;
  return pack;
} /* get_next_block */

/* Check filenames and phone numbers for 'bad' chars */
void check_name(char *name) {
  char *p;

  p = name;
  while (*p) {
    if (!(((*p <= 'Z') && (*p >= 'A'))
	  || ((*p <= 'z') && (*p >= 'a'))
	  || (*p == '.') || (*p == '_') || (*p == '@')
	  || ((*p <= '9') && (*p >= '0'))
	  || (*p == ' ') || (*p == ';')
	  || (*p == '/') || (*p == ':')))
      *p = '-';
    p++;
  }
} /* check_name */

int serial = 0;

/* Send TNPP packet via unix socket */
int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		      int len, int tnpp_to, int tnpp_from) {
  unsigned char buf[MAX_PACKET_SIZE];
  int length, crc;

  if (len > (MAX_PACKET_SIZE - 14))
    return;
  serial++;
  if (serial > 255)
    serial = 1;
  snprintf(buf, sizeof(buf) - 1, "%c%04X%02X%04X%02X%c", SOH, tnpp_to, 20,
	   tnpp_from, serial, STX);
  length = strlen(buf);
  memcpy(buf + strlen(buf), data, len);
  length = length + len;
  buf[length] = ETX;
  length++;
  crc = crc16(buf, length);
  buf[length] = (unsigned char) (crc & 0xff);
  buf[length + 1] = (unsigned char) ((crc >> 8) & 0xff);
  length = length + 2;

  /* Send packet now */
  return (sendto(fd, buf, length, 0, (struct sockaddr *) sun,
		 strlen(sun->sun_path) + 2));
} /* send_tnpp_packet */  

/* Receive tnpp packet from unix socket */
int recv_tnpp_packet(int fd, struct sockaddr_un *sun) {
  int res, len;
  unsigned char buf[MAX_PACKET_SIZE];
  char sbf[6];
  int dest, src;
  unsigned char *pblock;
  char funct;
  char id[21];
  unsigned char *pend;
  unsigned char *pxend;

  res = recv(fd, buf, sizeof(buf) - 1, 0);
  if (res < 0) {
    if ((errno != EINTR) && (errno != EAGAIN))
      wlog(L_WARNING, "Receive error %s (%d) at unix socket",
	   strerror(errno), errno);
    else
      wlog(L_DEBUG, "EINTR or EAGAIN received while receiving at unix socket");
    return -1;
  }
  len = res;
  buf[len] = ETX;
  sbf[4] = '\0';
  memcpy(sbf, buf + 1, 4);
  dest = strtol(sbf, NULL, 16);
  if (dest != my_tnpp_addr) {
    wlog(L_DEBUG, "Transit packet to destination %04X", dest);
    res = fork();
    if (res == 0) {
      sendto(fd, buf, len, 0, (struct sockaddr *) sun,
	     strlen(sun->sun_path) + 2);
      exit(0);
    } else
      if (res < 0)
	wlog(L_WARNING, "Fork error %s (%d)", strerror(errno), errno);
    return 0;
  }
  pblock = buf + 13;
  while ((*pblock != STX) && (pblock < (buf + len)))
    pblock++;
  if (*pblock != STX) {
    wlog(L_WARNING, "Bad format of TNPP packet (no STX)");
    return -1;
  }
  sbf[4] = '\0';
  memcpy(sbf, buf + 7, 4);
  src = strtol(sbf, NULL, 16);
  pend = pblock;
  do {
    pblock = pend + 1;
    if (*pblock == '>') {
      /* ETE request. Must generate ETE response */
      res = fork();
      if (res == 0) {
	sbf[0] = '<';
	sbf[1] = pblock[1];
	sbf[2] = pblock[2];
	if ((pblock[1] & 0xf0) != 7) {
	  /* No multiblock allowed */
	  sbf[3] = 'q';
	  sbf[4] = 'I';
	} else {
	  sbf[3] = 'a';
	  sbf[4] = '@';
	}
	sbf[5] = ETB;
	send_tnpp_packet(tnpp_out, sun, sbf, 6, src, my_tnpp_addr);
      } else
	if (res < 0)
	  wlog(L_WARNING, "Fork error %s (%d)", strerror(errno), errno);
      if ((pblock[1] & 0xf0) != 7) {
	wlog(L_WARNING, "Multiblock TNPP packet from %04X (rejected)", src);
	return -1;
      }
      pblock += 3;
    }
    for (pend = pblock + 1; (*pend != ETX) && (*pend != ETB); pend++);
    len = pend - pblock;
    if (len <= 1)
      continue;
    if ((*pblock != 'B') && (*pblock != 'C') && (*pblock != 'A')
	&& (*pblock != 'F')) {
      wlog(L_WARNING, "Unknown block type (%02X) in TNPP packet from %04X",
	   *pblock, src);
      continue;
    }
    if (*pblock == 'A') {
      quad_t pagenum;

      /* CAP block */
      if (len < 14) {
	wlog(L_WARNING, "Bad CAP Page format in packet from %04X", src);
	continue;
      }
      funct = *(pblock + 5);
      memcpy(id, pblock + 6, 8);
      id[8] = '\0';
      pagenum = strtoq(id, NULL, 16);
      sprintf(id, "!%qX", pagenum);
      pxend = strchr(id, '\0');
      pblock = pblock + 13;
      *pblock = CR;
      pblock = pblock - (pxend - (unsigned char *) id);
      len = len + (pxend - (unsigned char *) id) - 13;
      memcpy(pblock, id, pxend - (unsigned char *) id);
      wlog(L_INFO, "CAP page packet to '%s' (func %02X) from %04X, length %d",
	   id, funct, src, len);
      if (route_pack(id, pblock, len, funct))
	continue;
    } else if (*pblock == 'F') {
      int pgtype;
      int msn;
      int i;
      int usedlen;
      int caplen;
      quad_t pagenum;
      int msnlen;

      /* Extented CAP block */
      if (len < 15) {
	wlog(L_WARNING, "Bad Ext. CAP Page format in packet from %04X", src);
	continue;
      }
      pgtype = *(pblock + 1);
      msn = *(pblock + 6);
      funct = *(pblock + 5);
      caplen = 8;
      usedlen = 8;
      for (i = 0; i < CAPTYPES_NUMBER; i++)
	if (pgtype == captypes[i].pgtype) {
	  caplen = captypes[i].caplen;
	  usedlen = captypes[i].usedlen;
	  break;
	}
      if (pgtype == 0x66) {
	int pbl;

	/* Calculate CAP length for FLEX (tm) Augmented */
	if (*(pblock + 17) & 0x10)
	  caplen++;
	pbl = *(pblock + 18);
	if (pbl & 0x04)
	  caplen++;
	if (pbl & 0x02)
	  caplen++;
	if (pbl & 0x01)
	  caplen = caplen + 2;
      }
      if (msn & 0x20)
	msnlen = 2;
      else
	msnlen = 0;
      if (len < (7 + msnlen + caplen)) {
	wlog(L_WARNING, "Bad CAP Page format in packet from %04X", src);
	continue;
      }
      memcpy(id, pblock + 7, usedlen);
      id[usedlen] = '\0';
      pagenum = strtoq(id, NULL, 16);
      sprintf(id, "!%qX", pagenum);
      pxend = strchr(id, '\0');
      pblock = pblock + 6 + caplen + msnlen;
      *pblock = CR;
      pblock = pblock - (pxend - (unsigned char *) id);
      len = len + (pxend - (unsigned char *) id) - 6 - caplen - msnlen;
      memcpy(pblock, id, pxend - (unsigned char *) id);
      wlog(L_INFO, "Extended CAP page packet to '%s' (func %02X, ptype %02X, caplen %d, idlen %d) from %04X, length %d",
	   id, funct, pgtype, caplen, usedlen, src, len);
      if (route_pack(id, pblock, len, funct))
	continue;
    } else if (*pblock == 'C') {
      char cmdcode[4];
      char cmdmanf[4];
      char bufopt[512];
      char command[2048];

      /* COMMAND block */
      if (len < 7) {
	wlog(L_WARNING, "Bad format of COMMAND block in packet from %04X",
	     src);
	continue;
      }
      if (!command_filter) {
	wlog(L_WARNING, "No command to run when COMMAND block");
	continue;
      }
      memcpy(cmdmanf, pblock + 1, 3);
      memcpy(cmdcode, pblock + 4, 3);
      cmdmanf[3] = '\0';
      cmdcode[3] = '\0';
      if ((pend - pblock - 7) >= sizeof(bufopt)) {
	memcpy(bufopt, pblock + 7, sizeof(bufopt) - 1);
	bufopt[sizeof(bufopt) - 1] = '\0';
      } else {
	memcpy(bufopt, pblock + 7, pend - pblock - 7);
	bufopt[pend - pblock - 7] = '\0';
      }
      snprintf(command, sizeof(command), "%s '%s' '%s' '%s'",
	       command_filter, cmdmanf, cmdcode, bufopt);
      wlog(L_DEBUG, "Running command '%s'", command);
      res = fork();
      if (res == 0) {
	res = system(command);
	if (res != 0)
	  wlog(L_WARNING, "Command '%s' failed with code %d",
	       command, res);
	exit(res);
      } else
	if (res < 0)
	  wlog(L_WARNING, "Fork error %s (%d)", strerror(errno), errno);
    } else {
      int i;

      /* Page ID block */
      pblock++;
      funct = *pblock;
      pblock++;
      len = len - 2;
      if (len < 11) {
	wlog(L_WARNING, "Bad ID page packet format from %04X", src);
	continue;
      }
      for (i = 0; i < 10; i++)
	if (*(pblock + i) > ' ')
	  break;
      if (i < 10) {
	memcpy(id, pblock + i, 10 - i);
	id[10 - i] = '\0';
      } else
	id[0] = ' ';
      pblock += 10;
      len -= 10;
      pxend = strchr(id, ' ');
      if (pxend)
	*pxend = '\0';
      if ((char *) pxend == id) {
	wlog(L_WARNING, "Empty ID page packet format from %04X", src);
	continue;
      }
      if (!pxend)
	pxend = strchr(id, '\0');
      pblock--;
      *pblock = CR;
      pblock = pblock - (pxend - (unsigned char *) id);
      len = len + (pxend - (unsigned char *) id) + 1;
      memcpy(pblock, id, pxend - (unsigned char *) id);
      wlog(L_INFO, "ID page packet to '%s' (func %02X) from %04X, length %d",
	   id, funct, src, len);
      /*      wlog(L_DEBUG, "pxend '%s', id %s, data '%s'", pxend, id, pblock); */
      if (route_pack(id, pblock, len, funct))
	continue;
    }
  } while ((*pend != ETX) && (*(pend + 1) != ETX));
  return 0;
} /* recv_tnpp_packet */
 
/* SIGCHLD handler */
void childdead(int x) {
  int status;
  pid_t pid;

  while ((pid = wait3(&status, WNOHANG, NULL)) > 0)
    wlog(L_DEBUG, "Child %d has ended with status 0x%08x", pid, status);
  /*  signal(SIGCHLD, childdead); */
} /* childdead */

/* SIGHUP handler */
void sigreread(int x) {
  /* signal(SIGHUP, SIG_IGN); */
  wlog(L_WARNING, "SIGHUP has received");
  flag_reread = 1;
} /* sigreread */

/* SIGTERM handler */
void sigterm(int x) {
  wlog(L_WARNING, "SIGTERM or SIGINT has received");
  flag_fatal_error = 1;
} /* sigterm */

/* Search available device for destination (dest) */
tap_device *check_avail_dev(char *dest, int *max_r_len) {
  char *pphone;
  tap_device *dev;
  int i;
  char *cpar;
  char *cnext;
  char iphone[64];
  int maxlen;
  tap_device *dlist[2048];
  int dnum;

  dnum = 0;
  pphone = strchr(dest, ':');
  if (pphone)
    check_name(pphone + 1);
  for (dev = head.stqh_first; dev; dev = dev->devp.stqe_next) {
    for (i = 0; 1; i++) {
      if (dest[i] == '*') {
	if (dnum < 2048) {
	  wlog(L_DEBUG, "Found multi device %s for destination %s",
	       dev->name, dest);
	  dlist[dnum] = dev;
          dnum++;
	}
	break;
      }
      if (!dev->name[i]
	  && ((dest[i] == ':') || (dest[i] == ',') || !dest[i]))
	goto FOUND;
      else
	if (dev->name[i] != dest[i])
	  break;
    }
  }
  if (!dnum)
    return NULL;
  wlog(L_DEBUG, "Searching free device for dest %s from list of %d",
       dest, dnum);
  i = dlist[0]->next_multi;
  do {
    dev = dlist[i];
    wlog(L_DEBUG, "Trying device %s (%d) (phone %s, pphone %s)",
	 dev->name, i, dev->phone, pphone);
    if (pphone) {
      strncpy(iphone, pphone + 1, sizeof(iphone));
      cpar = strchr (pphone + 1, ';');
      if (cpar)
	iphone[cpar - (pphone + 1)] = '\0';
    }
    if ((dev->type != 0) || !dev->dial || !pphone
        || !dev->phone[0] || !strcmp(dev->phone, iphone)) {
      wlog(L_DEBUG, "Device %s (%d) is free. Next device is %d",
	   dev->name, i, (i + 1) % dnum);
      i = (i + 1) % dnum;
      dlist[0]->next_multi = i;
      goto FOUND;
    }
    if (dnum == 1)
      break;
    i = (i + 1) % dnum;
  } while (i != dlist[0]->next_multi);
  return NULL;
 FOUND:
  /* Device found. Check phone for dial-out devices */
  if ((dev->type != 0) || !dev->dial || !pphone)
    return dev;
  strncpy(iphone, pphone + 1, sizeof(iphone));
  /* Check for additional parameters for dial-out (max.len, max.pages) */
  cpar = strchr (pphone + 1, ';');
  if (cpar) {
    iphone[cpar - (pphone + 1)] = '\0';
    cpar++;
    if (*cpar && (*cpar != ';'))
      maxlen = atoi(cpar);
  } else
    maxlen = 999999;
  if (max_r_len)
    *max_r_len = maxlen;
  if (!dev->phone[0] || !strcmp(dev->phone, iphone)) {
    snprintf(dev->phone, sizeof(dev->phone) - 1, "%s", iphone);
    dev->max_r_len = 999999;
    dev->max_r_trans = 999999;
    if (cpar && *cpar) {
      cnext = strchr(cpar, ';');
      if (cnext != cpar)
	dev->max_r_len = atoi(cpar);
      if (cnext && *(cnext + 1))
	dev->max_r_trans = atoi(cnext + 1);
      if (dev->max_r_len <= 0) {
	wlog(L_WARNING, "Bad maximum length for device '%s' (phone '%s')",
	     dev->name, dev->phone);
	dev->max_r_len = 999999;
      }
      if (dev->max_r_trans <= 0) {
	wlog(L_WARNING, "Bad maximum number of transaction for device '%s' (phone '%s')",
	     dev->name, dev->phone);
	dev->max_r_trans = 999999;
      }
      wlog(L_DEBUG, "max_r_len = %d, max_r_trans = %d at device '%s' (phone '%s')",
	   dev->max_r_len, dev->max_r_trans, dev->name, dev->phone);
    }
    wlog(L_DEBUG, "check_avail_dev returns dial-up device %s", dev->name);
    return dev;
  }
  return NULL;
} /* check_avail_dev */

/* Search free device for packets in wo queue */
void check_wo_packs(tap_device *dev) {
  wo_packet *wpack;
  tap_device *fdev;

  for (wpack = wpacks.stqh_first; wpack; wpack = wpack->packp.stqe_next) {
    fdev = check_avail_dev(wpack->device, NULL);
    if (fdev) {
      wlog(L_DEBUG, "Found available device ('%s') for delayed packet",
	   fdev->name);
      if (fdev->npacks >= maxpacks)
	wlog(L_WARNING, "Queue full (too many packets) at device '%s'",
	     fdev->name);
      else {
	wpack->pack->nc = fdev->i_nc;
	STAILQ_INSERT_TAIL(&fdev->head, wpack->pack, packp);
	fdev->npacks++;
      }
      STAILQ_REMOVE(&wpacks, wpack, wo_packet, packp);
      free(wpack);
      nwpacks--;
    }
  }
} /* check_wo_packs */

/* Create uucp-style lock */
int uucp_lock(char *device) {
  char *p;
  int fd;
  char buf[128];
  char s[16];
  int res;

  for (p = device + strlen(device); (*p != '/') && (p != device); p--);
  snprintf(buf, sizeof(buf) - 1, "%s/LCK..%s", lock_dir, p + 1);
  fd = open(buf, O_CREAT | O_EXCL | O_WRONLY, 0666);
  if (fd < 0) {
    fd = open(buf, O_RDONLY);
    if (fd < 0)
      return -1;
    res = read(fd, s, sizeof(s));
    if (res > 0) {
      p = s;
      while (isspace(*p))
	p++;
      res = atoi(p);
      if (!kill(res, 0) && (res != getpid())) {
	wlog(L_DEBUG, "Lock file '%s' (pid %d) already exists", buf, res);
	return -1;
      }
    }
    close(fd);
    unlink(buf);
    fd = open(buf, O_EXCL | O_CREAT | O_WRONLY, 0666);
    if (fd < 0)
      return -1;
  }
  fchmod(fd, 0666);
  snprintf(s, sizeof(s) - 1, "% 10d\n", getpid());
  res = write(fd, s, strlen(s));
  close(fd);
  if (res != strlen(s))
    return -1;
  return 0;
} /* uucp_lock */

/* Remove uucp-style lock */
void uucp_unlock(char *device) {
  char *p;
  char buf[128];

  for (p = device + strlen(device); (*p != '/') && (p != device); p--);
  snprintf(buf, sizeof(buf) - 1, "%s/LCK..%s", lock_dir, p + 1);
  unlink(buf);
} /* uucp_unlock */

/* Create empty packet */
tap_packet *create_pack() {
  tap_packet *pack;

  pack = (tap_packet *) malloc(sizeof(tap_packet));
  if (pack) {
    pack->offset = pack->size = 0;
    pack->status = 0;
    pack->msg = NULL;
    pack->msglen = 0;
  }
  return pack;
} /* create_pack */

/* Put message-packet in head of queue */
void queue_ack(tap_device *dev, unsigned char *s) {
  tap_packet *pack;

  pack = create_pack();
  if (pack) {
    pack->size = strlen(s);
    memcpy(pack->data, s, pack->size);
    pack->status |= P_MESSAGE | P_SENDING;
    dev->status |= T_BLOCKWAIT;
    if (dev->lfcr || dev->crlf)
      conv_cr(pack, dev->crlf);
    STAILQ_INSERT_HEAD(&dev->head, pack, packp);
    dev->npacks++;
  }
  return;
} /* queue_ack */

/* Remove first packet from queue */
void dequeue_head(tap_device *dev) {
  tap_packet *pack;

  pack = dev->head.stqh_first;
  if ((dev->status & T_CALLOUT) || (dev->status & (T_INIT1 | T_INIT2)))
    dev->status &= ~(T_ACKWAIT | T_BLOCKWAIT);
  else
    dev->status &= ~T_ACKWAIT;
  if (pack) {
    STAILQ_REMOVE_HEAD(&dev->head, packp);
    free(pack);
    dev->npacks--;
    pack = dev->head.stqh_first;
    if (pack) {
      if (!(pack->status & P_WAITACK) && !(dev->status & T_IDENTWAIT)
	  && !(dev->status & T_CALLOUT)) {
	pack->status |= P_SENDING;
	dev->status |= T_BLOCKWAIT;
      }
    }
  }
  return;
} /* dequeue_head */

/* Remove all packets for first message in queue */
void dequeue_message(tap_device *dev) {
  tap_packet *pack;
  unsigned char *p;

  pack = dev->head.stqh_first;
  if (pack && (pack->status & P_DATA)) {
    for (p = pack->msg; pack
	   && (pack->msg == p); pack = dev->head.stqh_first)
      dequeue_head(dev);
    free(p);
  }
} /* dequeue_message */

/* Check call's limit for message */
void check_call_limit(tap_device *dev) {
  tap_packet *pack;

  pack = dev->head.stqh_first;
  if (pack && (pack->status & P_DATA)) {
    pack->nc--;
    if (!pack->nc) {
      wlog(L_WARNING, "Too many calls on device '%s'. Message dropped.",
	   dev->name);
      log_msg(pack->msg, pack->msglen, "too many unsuccessful calls");
      dequeue_message(dev);
      dev->status &= ~T_BLOCKWAIT;
    }
  }
} /* check_call_limit */

/* Put block's ack in queue */
void queue_pack_ack(tap_device *dev) {
  switch (dev->lastack) {
  case 0:
    queue_ack(dev, "211 Checksum is ok\r\06\r");
    break;
  case 1:
    queue_ack(dev, "514 Checksum error\r\025\r");
    break;
  case 2:
    queue_ack(dev, "515 Bad pager ID or message format\r\036\r");
    break;
  case 3:
    queue_ack(dev, "110 1.8\rHi there!\r\06\r\033[p\r");
    break;
  }
  return;
} /* queue_pack_ack */

/* Process new data */
void recv_data(tap_device *dev, int len) {
  unsigned char c;
  int i, j;
  int newpos;
  int flag;
  tap_packet *pack;
  unsigned char buf[32];
  char *p;
  int csum;
  int busy;
  unsigned char *msg;

  log_data("received", dev->buf + dev->offset, len);
 LOOP:
  wlog(L_DEBUG,"read_data (status 0x%04x, offset %d, len %d) at device '%s'",
       dev->status, dev->offset, len, dev->name);
  if (dev->status & T_CRWAIT) {
    for (i = dev->offset; i < (len + dev->offset); i++) {
      if ((dev->buf[i] & 0x7f) == CR) {
	/* Check for LF at end or begin of line */
	if ((i < (len + dev->offset)) && (dev->buf[i + 1] == LF)) {
	  if (i < (len + dev->offset - 1)) {
	    memmove(dev->buf + i + 1, dev->buf + i + 2,
		    len - (i - dev->offset) - 1);
	  }
	  len--;
	}
	if (dev->buf[dev->pos_line] == LF) {
	  memmove(dev->buf + dev->pos_line, dev->buf + dev->pos_line + 1,
		  len + (dev->offset - dev->pos_line) + 1);
	  i--;
	  len--;
	  dev->offset--;
	}
	if ((i > dev->offset) && (dev->buf[i - 1] == LF)) {
	  memmove(dev->buf + i - 1, dev->buf + i, len - (i - dev->offset));
	  i--;
	  len--;
	}
	goto FOUND_CR;
      }
    }
    dev->offset += len;
    return;
  } else
    if ((dev->status & T_CONNECTED) && (dev->status & T_CALLOUT)
	&& (dev->status & T_IDENTWAIT)) {
      dev->offset += len;
      if (strstr(dev->buf, "ID=")) {
	wlog(L_INFO, "ID request received at device '%s'", dev->name);
	if (dev->passwd)
	  snprintf(buf, sizeof(buf) - 1, "\033PG1%s\r", dev->passwd);
	else
	  snprintf(buf, sizeof(buf) - 1, "\033PG1\r");
	queue_ack(dev, buf);
	dev->status &= ~T_IDENTWAIT;
	dev->status |= T_CRWAIT;
	dev->offset = 0;
      }
      return;
    }
  wlog(L_ERROR, "Unexpected data in connected state (bug!) at device '%s'",
       dev->name);
  return;
 FOUND_CR:
  /* Found end-of-line */
  if (dev->status & T_CONNECTED) {
    if (dev->status & T_CALLOUT) {
      /* We're sending data */
      if (dev->status & T_ACKWAIT) {
	if (i == 0) {
	  newpos = i + 1;
	  goto SKIP_DATA;
	}
	c = dev->buf[i - 1];
	switch (c) {
	case ACK:
	  wlog(L_DEBUG, "ACK received at device '%s'", dev->name);
	  msg = dev->head.stqh_first->msg;
	  dequeue_head(dev);
	  dev->t3 = dev->i_t3;
	  dev->n2 = dev->i_n2;
	  dev->nr = dev->i_nr;
	  if (dev->head.stqh_first) {
	    dev->status |= T_BLOCKWAIT;
	    if (msg != dev->head.stqh_first->msg) {
	      free(msg);
	      dev->n_trans++;
	    }
	  } else {
	    free(msg);
	    dev->n_trans++;
	  }
	  dev->status &= ~T_ACKWAIT;
	  break;
	case NAK:
	  wlog(L_DEBUG, "NAK received at device '%s'", dev->name);
	  dev->t3 = dev->i_t3;
	  dev->n2 = dev->i_n2;
	  dev->nr--;
	  if (dev->nr <= 0) {
	    wlog(L_WARNING, "Too many NAK at device '%s'", dev->name);
	    msg = dev->head.stqh_first->msg;
	    log_msg(msg, dev->head.stqh_first->msglen, "too many NAKs");
	    dequeue_message(dev);
	    write(dev->fd, "\04\r", 2);
	    dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
	    dev->status |= T_CALLEND;
	    dev->offset = dev->pos_line = dev->pos_block = 0;
	    return;
	  }
	  pack = dev->head.stqh_first;
	  pack->status &= ~P_WAITACK;
	  pack->status |= P_SENDING;
	  pack->offset = 0;
	  dev->status &= ~T_ACKWAIT;
	  dev->status |= T_BLOCKWAIT;
	  break;
	case RS:
	  wlog(L_DEBUG, "RS received at device '%s'", dev->name);
	  log_msg(dev->head.stqh_first->msg, dev->head.stqh_first->msglen,
		  "RS received");
	  dequeue_message(dev);
	  dev->t3 = dev->i_t3;
	  dev->n2 = dev->i_n2;
	  dev->nr = dev->i_nr;
	  if (dev->head.stqh_first)
	    dev->status |= T_BLOCKWAIT;
	  dev->status &= ~T_ACKWAIT;
	  dev->n_trans++;
	  break;
	default:
	  break;
	}
	if (!dev->head.stqh_first || (dev->n_trans >= dev->max_r_trans)
	    || (dev->n_trans >= dev->max_trans)) {
	  dev->status &= ~(T_BLOCKWAIT | T_CRWAIT);
	  dev->status |= T_CALLEND;
	  write(dev->fd, "\04\r", 2);
	}
	newpos = i + 1;
	goto SKIP_DATA;
      } else
	if (!(dev->status & T_BLOCKWAIT)) {
	  /* Waiting for greetings */
	  if (i == 0) {
	    newpos = i + 1;
	    goto SKIP_DATA;
	  }
	  c = dev->buf[i - 1];
	  switch (c) {
	  case ACK:
	    break;
	  case NAK:
	    dev->n1--;
	    if (dev->n1 <= 0) {
	      wlog(L_INFO, "Unsuccessful login (too many NAKs) at device '%s'",
		   dev->name);
	      write(dev->fd, "\04\r", 2);
	      dev->status &= ~T_CRWAIT;
	      dev->status |= T_CALLEND;
	      dev->offset = dev->pos_line = dev->pos_block = 0;
	      check_call_limit(dev);
	      return;
	    }
	    if (dev->passwd)
	      snprintf(buf, sizeof(buf) - 1, "\033PG1%s\r", dev->passwd);
	    else
	      snprintf(buf, sizeof(buf) - 1, "\033PG1\r");
	    queue_ack(dev, buf);
	    break;
	  case EOT:
	    wlog(L_INFO, "Unsuccessful login (EOT) at device '%s'",
		 dev->name);
	    write(dev->fd, "\04\r", 2);
	    dev->status &= ~T_CRWAIT;
	    dev->status |= T_CALLEND;
	    dev->offset = dev->pos_line = dev->pos_block = 0;
	    check_call_limit(dev);
	    return;
	  case 'p':
	    if (i < 3) {
	      newpos = i + 1;
	      goto SKIP_DATA;
	    }
	    if ((dev->buf[i - 2] == '[') && (dev->buf[i - 3] == ESC)) {
	      wlog(L_INFO, "Starting transmission at device '%s'",
		   dev->name);
	      if (dev->head.stqh_first)
		dev->status |= T_BLOCKWAIT;
	      else {
		wlog(L_WARNING, "No data (why we calling?) at device '%s'",
		     dev->name);
		dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
		dev->status |= T_CALLEND;
		dev->offset = dev->pos_line = dev->pos_block = 0;
		return;
	      }
	    }
	    break;
	  default:
	    break;
	  }
	  newpos = i + 1;
	  goto SKIP_DATA;
	}
    } else {
      /* We're receiving data */
      if (dev->status & T_IDENTWAIT) {
	if (dev->status & T_ACKWAIT) {
	  /* Check parity */
	  if (dev->autodetect) {
	    if (dev->buf[i] == 0x8d) {
	      wlog(L_DEBUG, "Parity is ON at device '%s'", dev->name);
	      set_port_parity(dev->fd, 7, 1);
	    } else {
	      wlog(L_DEBUG, "Parity is OFF at device '%s'", dev->name);
	      set_port_parity(dev->fd, 7, 0);
	    }
	  }
	  /* Send ID-request after CR */
	  dev->status &= ~T_ACKWAIT;
	  queue_ack(dev, "ID=");
	  newpos = i + 1;
	  goto SKIP_DATA;
	}
	/* Check for identify string (PG1 or M) */
	wlog(L_DEBUG, "Looking for ID string at device '%s'", dev->name);
	if (i == dev->pos_line) {
	  newpos = i + 1;
	  goto SKIP_DATA;
	}
	if (!(dev->buf[dev->pos_line] == 'M')
	    && !(!strncasecmp(dev->buf + dev->pos_line, "PG1", 3)
		 || !strncasecmp(dev->buf + dev->pos_line + 1, "PG1", 3)))
	  wlog(L_WARNING, "Unknown remote type at device '%s'", dev->name);
	dev->status &= ~T_IDENTWAIT;
	dev->status |= T_BLOCKWAIT;
	newpos = i + 1;
	dev->lastack = 3;
	queue_pack_ack(dev);
	goto SKIP_DATA;
      } else {
	if (dev->status & T_BLOCKWAIT) {
	  wlog(L_DEBUG, "Checking for new block at device '%s'", dev->name);
	  if (dev->buf[dev->pos_line] != STX) {
	    if (dev->buf[dev->pos_line] == EOT) {
	      wlog(L_INFO, "Disconnect sequence at device '%s'", dev->name);
	      dev->status &= ~(T_BLOCKWAIT | T_CRWAIT);
	      dev->status |= T_CALLEND;
	      write(dev->fd, "115 Bye-bye!\r\033\04\r", 16);
	      dev->offset = dev->pos_line = dev->pos_block = 0;
	      return;
	    }
	    wlog(L_INFO, "Junk before block at device '%s'",
		 dev->name);
	    queue_ack(dev, "502 Start transmission, please\r\036\r");
	    newpos = i + 1;
	    goto SKIP_DATA;
	  }
	  /* New block! */
	  dev->t4 = dev->i_t4;
	  dev->n2 = dev->i_n2;
	  memmove(dev->buf + dev->pos_line,
		  dev->buf + dev->pos_line + 1,
		  dev->offset + len - dev->pos_line - 1);
	  dev->pos_block = dev->pos_line;
	  len--;
	  i--;
	  dev->status &= ~T_BLOCKWAIT;
	  dev->status |= T_BLOCKEND;
	}
	if (dev->status & T_BLOCKEND) {
	  wlog(L_DEBUG, "Checking for end of block at device '%s'",
	       dev->name);
	  if ((i - dev->pos_line) < 4) {
	    newpos = i + 1;
	    goto NEXT_DATA;
	  }
	  c = dev->buf[i - 4];
	  if ((c != US) && (c != ETB) && (c != ETX)) {
	    newpos = i + 1;
	    goto NEXT_DATA;
	  }
	  /* End of block */
	  dev->status &= ~T_BLOCKEND;
	  dev->status |= T_BLOCKWAIT;
	  /* Verify checksum */
	  csum = (((dev->buf[i - 3] - 48) << 8)
		  | ((dev->buf[i - 2] - 48) << 4)
		  | (dev->buf[i - 1] - 48)) - 2;
	  if (checksum(dev->buf + dev->pos_block,
		       i - dev->pos_block - 3) != (csum & 0x0fff)) {
	    wlog(L_INFO, "Bad checksum of block at device '%s'", dev->name);
	    dev->lastack = 1;
	    queue_pack_ack(dev);
	    newpos = i + 1;
	    goto SKIP_DATA;
	  }
	  wlog(L_INFO, "New block was received at device '%s'", dev->name);
	  if (c != ETX) {
	    dev->lastack = 0;
	    queue_pack_ack(dev);
	    newpos = i - 4;
	    dev->pos_block = newpos;
	    len = len - 5;
	    goto NEXT_DATA;
	  } else {
	    /* End of transaction */
	    if (process_tap_packet(dev->buf, i - 4)) {
	      dev->lastack = 0;
	      queue_pack_ack(dev);
	    } else {
	      wlog(L_WARNING, "Could not process message at device '%s'",
		   dev->name);
	      dev->lastack = 2;
	      queue_pack_ack(dev);
	    }
	    dev->pos_block = dev->pos_line = dev->offset = 0;
	    dev->status &= ~T_BLOCKEND;
	    dev->status |= T_BLOCKWAIT | T_CRWAIT;
	    return;
	  }
	}
      }
    }
  } else {
    /* Not connected */
    if (dev->status & (T_INIT1 | T_INIT2)) {
      if (i > 1) {
	if ((dev->buf[i - 2] == 'O') && (dev->buf[i - 1] == 'K')) {
	  if ((dev->status & T_INIT1) && dev->init2 && dev->init2[0]) {
	    queue_ack(dev, dev->init2);
	    dev->status &= ~T_INIT1;
	    dev->status |= T_INIT2 | T_CRWAIT;
	  } else {
	    dev->status &= ~(T_INIT1 | T_INIT2);
	    uucp_unlock(dev->device);
	    wlog(L_INFO, "Port unlocked on device '%s'", dev->name);
	    unblock_concurrent(dev);
	    if (dev->lastdial) {
		dev->tblock = dev->i_tblock;
		dev->lastdial = 1;
	    }
	  }
	} else if (!((dev->buf[0] == 'A') && (dev->buf[1] == 'T'))
		   && !((dev->buf[0] == 'N') && (dev->buf[1] == 'O'))) {
	  wlog(L_WARNING, "Can't init modem at device '%s'", dev->name);
	  dev->status &= ~T_INIT2;
	  dev->status |= T_INIT1 | T_CRWAIT;
	  queue_ack(dev, dev->init1);
	  dev->tblock = dev->i_tblock;
	  uucp_unlock(dev->device);
	  wlog(L_INFO, "Port unlocked on device '%s'", dev->name);
	  unblock_concurrent(dev);
	} else if ((dev->status & T_INIT1)
		   &&(dev->buf[0] == 'N') && (dev->buf[1] == 'O'))
	  queue_ack(dev, dev->init1);
      }
      newpos = i + 1;
      goto SKIP_DATA;
    }
    if ((dev->status & T_CALLOUT) && (dev->status & T_DIALSTATE)) {
      if (!strncmp(dev->buf + dev->pos_line, "CONNECT", 7)) {
	dev->buf[i] = '\0';
	if (i > (dev->pos_line + 8))
	  wlog(L_INFO, "Connect '%s' at device '%s'",
	       dev->buf + dev->pos_line + 8, dev->name);
	else
	  wlog(L_INFO, "Connect (300 baud probably) at device '%s'",
	       dev->name);
      } else
	if (!strncmp(dev->buf + dev->pos_line, "NO", 2)
	    || !(busy = strncmp(dev->buf + dev->pos_line, "BUSY", 4))) {
	  dev->buf[i] = '\0';
	  wlog(L_INFO, "Dialing error '%s' at device '%s'",
	       dev->buf + dev->pos_line, dev->name);
	  dev->status |= T_CALLEND;
	  dev->lastdial = 1;
	  if (!busy)
	    check_call_limit(dev);
	}
    }
    newpos = i + 1;
    goto SKIP_DATA;
  }
  wlog(L_DEBUG, "Nothing to do at device '%s'", dev->name);
  return;

 SKIP_DATA:
  /* Drop unused chars or last block */
  dev->status |= T_CRWAIT;
  dev->pos_line = dev->pos_block;
  if ((newpos - dev->offset) < len) {
    /* We have more data to processing... */
    len = len - (newpos - dev->offset);
    dev->offset = dev->pos_block;
    memmove(dev->buf + dev->pos_block, dev->buf + newpos, len);
    goto LOOP;
  }
  dev->offset = dev->pos_block;
  return;

 NEXT_DATA:
  /* Process (or wait for) next line */
  dev->pos_line = newpos;
  dev->status |= T_CRWAIT;
  if ((newpos - dev->offset) < len) {
    /* Hey! There some data in buffer! */
    len = len - (newpos - dev->offset);
    dev->offset = newpos;
    goto LOOP;
  }
  dev->offset = newpos;
  return;
} /* recv_data */

/* Prepare new data to send */
void send_data(tap_device *dev, int len) {
  tap_packet *pack;

  pack = dev->head.stqh_first;
  log_data("sent", pack->data + pack->offset, len);
  pack->offset += len;
  wlog(L_DEBUG,"send_data (status 0x%04x, len %d) at device '%s'",
       dev->status, len, dev->name);
  if (pack->offset < pack->size)
    return;
  /* Packet is done */
  if (pack->status & P_DATA) {
    /* If data packet - wait for ack */
    dev->status &= ~T_BLOCKWAIT;
    dev->status |= T_ACKWAIT;
    pack->status &= ~P_SENDING;
    pack->status |= P_WAITACK;
    return;
  }
  /* Drop non-data packet after sending */
  dequeue_head(dev);
  if (dev->head.stqh_first) {
    dev->head.stqh_first->offset = 0;
    dev->head.stqh_first->status &= ~P_WAITACK;
    dev->head.stqh_first->status |= P_SENDING;
  }
  return;
} /* send_data */

void DTR_up(int fd) {
  int flag;

  flag = TIOCM_DTR;
  ioctl(fd, TIOCMBIS, &flag);
} /* DTR_up */

void DTR_down(int fd) {
  int flag;

  flag = TIOCM_DTR;
  ioctl(fd, TIOCMBIC, &flag);
} /* DTR_down */

void toggle_dtr(int fd) {
  struct timeval tp;

  DTR_down(fd);
  tp.tv_sec = 0;
  tp.tv_usec = 250000;
  select(0, NULL, NULL, NULL, &tp);
  DTR_up(fd);
  tp.tv_sec = 0;
  tp.tv_usec = 250000;
  select(0, NULL, NULL, NULL, &tp);
} /* toggle_dtr */

/* Block all concurrent devices */
void block_concurrent(tap_device *dev) {
  tap_device *tdev;
  tap_device *block_dev;

  if (dev->master) {
    block_dev = dev->master;
    block_dev->status |= T_BUSY;
  } else
    block_dev = dev;
  for (tdev = head.stqh_first; tdev;
       tdev = tdev->devp.stqe_next)
    if ((tdev->master == block_dev) && (tdev != dev))
      tdev->status |= T_BUSY;
} /* block_concurrent */

/* Unblock all concurrent devices */
void unblock_concurrent(tap_device *dev) {
  tap_device *tdev;
  tap_device *block_dev;

  if (dev->master) {
    block_dev = dev->master;
    block_dev->status &= ~T_BUSY;
  } else
    block_dev = dev;
  for (tdev = head.stqh_first; tdev;
       tdev = tdev->devp.stqe_next)
    if ((tdev->master == block_dev) && (tdev != dev))
      tdev->status &= ~T_BUSY;
} /* unblock_concurrent */

/* main loop */
void loop() {
  tap_device *dev;
  int i;
  int res;
  tap_packet *pack;
  fd_set rfd, wfd, efd;
  unsigned char buf[256];
  int fd_max;
  time_t b_t, a_t;
  struct timeval tp;
  tap_device *tdev;
  struct sigaction sact;

  flag_fatal_error = 0;
  flag_reread = 0;
  wlog(L_INFO, "Entering into main loop");
  FD_ZERO(&rfd);
  FD_ZERO(&wfd);
  FD_ZERO(&efd);
  sact.sa_handler = &childdead;
  sact.sa_flags = SA_NODEFER | SA_NOCLDSTOP;
  sigemptyset(&sact.sa_mask);
  sigaction(SIGCHLD, &sact, NULL);
  sact.sa_handler = &sigreread;
  sact.sa_flags = 0;
  sigaction(SIGHUP, &sact, NULL);
  sact.sa_handler = &sigterm;
  sigaction(SIGINT, &sact, NULL);
  sigaction(SIGTERM, &sact, NULL);
  /* loop */
  do {
    fd_max = tnpp_in;
    FD_SET(tnpp_in, &rfd);
    for (dev = head.stqh_first; dev; dev = dev->devp.stqe_next) {
      if (!dev->tblock && (dev->type == 0)) {
	FD_SET(dev->fd, &efd);
	if (fd_max < dev->fd)
	  fd_max = dev->fd;
	if (!(dev->init1 && dev->init1[0]) && (dev->status & T_INIT1)) {
	  dev->status &= ~T_INIT1;
	  unblock_concurrent(dev);
	}
	if (((dev->status & (T_INIT1 | T_INIT2 | T_DIALSTATE))
	     && dev->head.stqh_first
	     && (dev->head.stqh_first->status & P_MESSAGE))
	    || (dev->head.stqh_first && (dev->status & T_CONNECTED)
		&& (!(dev->status & T_CALLOUT) || (dev->status & T_BLOCKWAIT))
		&& !((dev->head.stqh_first->status & P_DATA)
		     && (dev->status & T_ACKWAIT))))
	  FD_SET(dev->fd, &wfd);
	else
	  if (!(dev->status & T_ACKWAIT) && dev->head.stqh_first
	      && (dev->head.stqh_first->status & P_DATA)) {
	    if ((dev->clocal
		 && !(dev->status & (T_DIALSTATE | T_CALLEND | T_INIT1
				     | T_INIT2 | T_IDENTWAIT | T_BLOCKWAIT
				     | T_BLOCKEND | T_CRWAIT | T_BUSY)))
		|| (!dev->clocal && !(dev->status & (T_CONNECTED | T_INIT1
				      | T_INIT2 | T_DIALSTATE | T_BUSY)))) {
	      /* Outgoing call */
	      if (dev->type == 0) {
		res = uucp_lock(dev->device);
		if (!res) {
		  wlog(L_INFO, "Port locked on device '%s'",
		       dev->name);
		  set_serial(dev);
		  DTR_up(dev->fd);
		  tcflush(dev->fd, TCIOFLUSH);
		  dev->status = T_CALLOUT | T_CRWAIT | T_DIALSTATE;
		  if (dev->dial && !dev->clocal) {
		    snprintf(buf, sizeof(buf) - 1, "%s%s%s",
			     dev->dial_p, dev->phone, dev->dial_s);
		    queue_ack(dev, buf);
		    FD_SET(dev->fd, &wfd);
		  }
		  block_concurrent(dev);
		  dev->tsess = dev->i_tsess;
		} else {
		  wlog(L_INFO, "Could not lock port '%s' (%s) for device '%s'",
		       dev->device, strerror(errno), dev->name);
		  dev->status |= T_INIT1 | T_CRWAIT;
		  if (dev->init1 && dev->init1[0])
		    queue_ack(dev, dev->init1);
		  dev->tblock = dev->i_tblock;
		}
	      }
	    }
	  }
	if ((dev->type == 0) && !(dev->status & (T_CONNECTED | T_INIT1 | T_BUSY
						 | T_INIT2| T_CALLEND))) {
	  /* Check CD for serial ports */
	  res = 0;
	  if (!dev->clocal && (!dev->no_incoming
			       || (dev->status & T_CALLOUT)))
	    ioctl(dev->fd, TIOCMGET, &res);
	  if ((dev->clocal && (dev->status & T_CALLOUT)
	       && (dev->status & T_DIALSTATE))
	      || ((res & TIOCM_CD)
                  && ((dev->status & T_CALLOUT) || !dev->master))) {
	    wlog(L_INFO, "Remote connected to device '%s'",
		 dev->name);
	    if (!(dev->status & T_DIALSTATE)) {
	      /* Incoming call */
	      if (!dev->clocal && !(dev->status & T_CALLOUT))
		strcpy(dev->phone, "Incoming");
	      res = uucp_lock(dev->device);
	      if (!res) {
		wlog(L_INFO, "Port locked on device '%s'",
		     dev->name);
		set_serial(dev);
		block_concurrent(dev);
	      } else {
		wlog(L_INFO, "Could not lock port '%s' (%s) for device '%s'",
		     dev->device, strerror(errno), dev->name);
		dev->status |= T_INIT1 | T_CRWAIT;
		if (dev->init1 && dev->init1[0])
		  queue_ack(dev, dev->init1);
		dev->tblock = dev->i_tblock;
		goto CHECK_RFD;
	      }
	    } else
	      dev->n_trans = 0;
	    DTR_up(dev->fd);
	    tcflush(dev->fd, TCIOFLUSH);
	    /* device to connected state */
	    dev->status &= ~(T_DIALSTATE | T_CRWAIT);
	    dev->status |= T_CONNECTED | T_IDENTWAIT;
	    dev->tr = dev->i_tr;
	    dev->t1 = dev->i_t1;
	    dev->t2 = dev->i_t2;
	    dev->t3 = dev->i_t3;
	    dev->t4 = dev->i_t4;
	    dev->t5 = dev->i_t5;
	    dev->n1 = dev->i_n1;
	    dev->n2 = dev->i_n2;
	    dev->n3 = dev->i_n3;
	    dev->nr = dev->i_nr;
	    dev->pos_line = dev->offset = dev->pos_block = 0;
	    dev->tsess = dev->i_tsess;
	    if (dev->status & T_CALLOUT) {
	      queue_ack(dev, "\r");
	      FD_SET(dev->fd, &wfd);
	    } else {
	      dev->status |= (T_CRWAIT | T_IDENTWAIT | T_ACKWAIT);
	      if (dev->autodetect)
		set_port_parity(dev->fd, 8, 0);
	    }
	  }
	}
      CHECK_RFD:
	if ((dev->status & (T_INIT1 | T_INIT2)) ||
	    ((dev->status & T_CONNECTED) && !(dev->status & T_CALLEND)) ||
	    (dev->status & T_DIALSTATE))
	  /* We should read data */
	  FD_SET(dev->fd, &rfd);
      }
    }
    time(&b_t);
    tp.tv_sec = 1;
    tp.tv_usec = 0;
    /* Select */
    res = select(fd_max + 1, &rfd, &wfd, &efd, &tp);
    wlog(L_DEBUG, "Select returned %d", res);
    if (res < 0) {
      if (errno != EINTR) {
	wlog(L_ERROR, "Error %s (%d) at select. Exiting",
	     strerror(errno), errno);
	flag_fatal_error = 1;
	break;
      }
      wlog(L_DEBUG, "EINTR at select");
      res = 0;
      FD_ZERO(&rfd);
      FD_ZERO(&efd);
      FD_ZERO(&wfd);
      if (flag_fatal_error || flag_reread)
	break;
    }
    for (dev = head.stqh_first;dev; dev = dev->devp.stqe_next) {
      if (dev->status & T_BUSY)
        continue;
      if ((dev->status & T_CONNECTED) && !dev->clocal) {
	ioctl(dev->fd, TIOCMGET, &res);
	if (!(res & TIOCM_CD)) {
	  wlog(L_INFO, "Remote disconnected from device '%s'",
	       dev->name);
	  dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
	  dev->status |= T_CALLEND;
	  dev->offset = dev->pos_line = dev->pos_block = 0;
	  FD_CLR(dev->fd, &rfd);
	  FD_CLR(dev->fd, &wfd);
	}
      }
      if ((dev->fd >= 0) && FD_ISSET(dev->fd, &rfd)) {
	FD_CLR(dev->fd, &rfd);
	/* Read from device */
	if (dev->clocal && !(dev->status & (T_CALLOUT | T_BLOCKWAIT
					    | T_IDENTWAIT | T_ACKWAIT
					    | T_CALLEND | T_BLOCKEND))
	    && !dev->master) {
	  /* We cannot detect new connections on clocal devices by
	   * other way :( */
	  wlog(L_INFO, "New connection probably at device '%s'",
	       dev->name);
	  dev->status |= T_IDENTWAIT | T_ACKWAIT | T_CRWAIT | T_CONNECTED;
	  block_concurrent(dev);
	  dev->tr = dev->i_tr;
	  dev->t1 = dev->i_t1;
	  dev->t2 = dev->i_t2;
	  dev->t3 = dev->i_t3;
	  dev->t4 = dev->i_t4;
	  dev->t5 = dev->i_t5;
	  dev->n1 = dev->i_n1;
	  dev->n2 = dev->i_n2;
	  dev->n3 = dev->i_n3;
	  dev->nr = dev->i_nr;
	  dev->offset = dev->pos_line = dev->pos_block = 0;
	  dev->tsess = dev->i_tsess;
	  if (dev->autodetect)
	    set_port_parity(dev->fd, 8, 0);
	}
	if (dev->offset >= MAX_PACKET_SIZE) {
	  wlog(L_WARNING, "Too big packet at device '%s'",
	       dev->name);
	  dev->offset = dev->pos_line = dev->pos_block = 0;
	}
	res = read(dev->fd, dev->buf + dev->offset,
		   MAX_PACKET_SIZE - dev->offset);
	if (res < 0) {
	  if ((errno != EINTR) && (errno != EAGAIN)) {
	    wlog(L_WARNING, "Read error %s (%d) at device %s",
		 strerror(errno), errno, dev->name);
	    if (dev->type)
	      reconnect(dev);
	    continue;
	  }
	} else
	  if (res > 0) {
	    dev->tr = dev->i_tr;
	    recv_data(dev, res);
	  }
      }
      if ((dev->fd >= 0) && FD_ISSET(dev->fd, &wfd)) {
	FD_CLR(dev->fd, &wfd);
	/* Send first block from queue */
	if ((pack = dev->head.stqh_first) && !(pack->status & P_WAITACK)
	    && (pack->offset != pack->size)) {
	  if ((dev->status & T_INIT1) && (dev->type == 0)
	      && dev->dtrtoggle && dev->head.stqh_first
	      && (dev->head.stqh_first->status & P_MESSAGE)) {
	    /* Toggle DTR before send init string to modem */
	    toggle_dtr(dev->fd);
	    tcflush(dev->fd, TCIOFLUSH);
	  }
	  res = write(dev->fd, pack->data + pack->offset,
		      pack->size - pack->offset);
	  if (res < 0) {
	    if ((errno != EINTR) && (errno != EAGAIN)) {
	      wlog(L_WARNING, "Write error %s (%d) at device '%s'",
		   strerror(errno), errno, dev->name);
	      if (dev->type == 0)
		reconnect(dev);
	      continue;
	    }
	  } else
	    if (res > 0) {
	      dev->tr = dev->i_tr;
	      send_data(dev, res);
	    }
	} else
	  wlog(L_WARNING, "No data to send (bug!) at device '%s'",
	       dev->name);
      }
      /* Exceptions checking */
      if (!flag_fatal_error && (dev->fd >= 0) && FD_ISSET(dev->fd, &efd)) {
	FD_CLR(dev->fd, &efd);
	wlog(L_WARNING, "Exception %s (%d) at device '%s'",
	     strerror(errno), errno, dev->name);
	if (dev->type == 0)
	  reconnect(dev);
      }
      if (dev->status & T_CALLEND) {
	/* End of call */
	if (dev->autodetect) {
	  if (dev->parity)
	    set_port_parity(dev->fd, 7, 1);
	  else
	    set_port_parity(dev->fd, 7, 0);
	}
	dev->offset = dev->pos_line = dev->pos_block = 0;
	unblock_concurrent(dev);
	/* Remove first packets from queue, if it is not data packets */
	while (dev->head.stqh_first
	       && (dev->head.stqh_first->status & P_MESSAGE))
	  dequeue_head(dev);
	if (dev->head.stqh_first) {
	  dev->head.stqh_first->status &= ~(P_SENDING | P_WAITACK);
	  dev->head.stqh_first->offset = 0;
	} else {
	  dev->phone[0] = '\0';
	  check_wo_packs(dev);
	}
	if (dev->master) {
	  tdev = dev->master;
          dev->status = 0;
	} else
	  tdev = dev;
	tdev->tblock = 2;
	reconnect(dev);
	if (tdev->init1 && tdev->init1[0]) {
	  /* Init modem with init1 after one-second delay */
	  if (!tdev->clocal) {
	    block_concurrent(tdev);
	    tdev->status = T_INIT1 | T_CRWAIT;
	    queue_ack(tdev, tdev->init1);
	  } else {
	    if (tdev->dtrtoggle) {
	      DTR_down(tdev->fd);
	    }
	    tdev->status &= ~(T_CALLEND | T_CALLOUT);
	  }
	} else {
	  /* Hmmm, no init string */
	  if (tdev->dtrtoggle) {
	    if (tdev->clocal || !tdev->dial)
	      DTR_down(tdev->fd);
	    else
	      toggle_dtr(tdev->fd);
	  }
	  tdev->status = 0;
	}
	uucp_unlock(dev->device);
	wlog(L_INFO, "Port unlocked on device '%s'", dev->name);
      }
    }
    if (FD_ISSET(tnpp_in, &rfd)) {
      FD_CLR(tnpp_in, &rfd);
      /* Data packet from rtnppd */
      recv_tnpp_packet(tnpp_in, &u_sun);
    }
    time(&a_t);
    if (a_t > b_t) {
      /* Check timers */
      for (dev = head.stqh_first; dev; dev = dev->devp.stqe_next) {
	if (dev->status & T_BUSY)
	  continue;
	if (dev->type == 0) {
	  if (dev->tblock) {
	    /* Unblock device? */
	    dev->tblock--;
	    if ((dev->tblock == 0) && (dev->status & (T_INIT1 | T_INIT2)) &&
		(dev->init1 && dev->init1[0])) {
	      res = uucp_lock(dev->device);
	      if (!res) {
		wlog(L_INFO, "Port locked on device '%s'",
		     dev->name);
		set_serial(dev);
		tcflush(dev->fd, TCIOFLUSH);
		block_concurrent(dev);
	      } else {
		wlog(L_INFO, "Could not lock port '%s' (%s) for device '%s'",
		     dev->device, strerror(errno), dev->name);
		dev->tblock = dev->i_tblock;
		unblock_concurrent(dev);
	      }
	    }
	  }
	  if ((dev->status & T_CONNECTED) && !(dev->status & T_CALLEND)
	      && !dev->clocal) {
	    dev->tsess--;
	    if (dev->tsess <= 0) {
	      wlog(L_WARNING, "Max session time reached on device '%s'",
		   dev->device);
	      dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
	      dev->status |= T_CALLEND;
	      dev->offset = dev->pos_line = dev->pos_block = 0;
	      dev->tr = dev->i_tr;
	    }
	  }
	}
	if (!(dev->status & T_CALLEND)) {
	  if (dev->status & (T_INIT1 | T_INIT2 | T_BLOCKWAIT | T_IDENTWAIT
			     | T_ACKWAIT | T_DIALSTATE)
	      && !dev->tblock) {
	    /* Keep-alive check */
	    dev->tr--;
	    if (dev->tr == 0) {
	      wlog(L_WARNING, "Timeout at device '%s'", dev->name);
	      dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
	      dev->status |= T_CALLEND;
	      dev->offset = dev->pos_line = dev->pos_block = 0;
	      dev->tr = dev->i_tr;
	    }
	  }
	  if ((dev->status & T_CONNECTED) && (dev->status & T_IDENTWAIT)
	      && !(dev->status & T_ACKWAIT) && (dev->status & T_CALLOUT)) {
	    /* Wait for 'ID=' */
	    dev->t1--;
	    if (dev->t1 == 0) {
	      wlog(L_DEBUG, "Identify string wait timeout at device '%s'",
		   dev->name);
	      dev->n1--;
	      if (dev->n1 == 0) {
		wlog(L_WARNING, "Could not get identify string at device '%s'",
		     dev->name);
		dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
		dev->status |= T_CALLEND;
		dev->offset = dev->pos_line = dev->pos_block = 0;
		dev->n1 = dev->i_n1;
	      } else
		queue_ack(dev, "\r");
	      dev->t1 = dev->i_t1;
	    }
	  }
	  if ((dev->status & T_CONNECTED) && (dev->status & T_IDENTWAIT)
	      && !(dev->status & T_CALLOUT)) {
	    /* Wait for 'PG1' or 'M' */
	    dev->t5--;
	    if (dev->t5 == 0) {
	      wlog(L_DEBUG, "PG1 string wait timeout at device '%s'",
		   dev->name);
	      dev->n3--;
	      if (dev->n3 == 0) {
		wlog(L_WARNING, "Could not get PG1 string at device '%s'",
		     dev->name);
		dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
		dev->status |= T_CALLEND;
		dev->offset = dev->pos_line = dev->pos_block = 0;
		dev->n3 = dev->i_n3;
	      } else {
		/* Set parity to default if no CR received before send ID= */
		if (dev->autodetect) {
		  if (dev->parity)
		    set_port_parity(dev->fd, 7, 1);
		  else
		    set_port_parity(dev->fd, 7, 0);
		}
		queue_ack(dev, "ID=");
		dev->status &= ~T_ACKWAIT;
	      }
	      dev->t5 = dev->i_t5;
	    }
	  }
	  if ((dev->status & T_CALLOUT) && !(dev->status & T_IDENTWAIT)
	      && (dev->status & T_ACKWAIT) && (dev->status & T_CONNECTED)) {
	    /* Wait for ACK */
	    dev->t3--;
	    if (dev->t3 == 0) {
	      wlog(L_DEBUG, "ACK wait timeout at device '%s'", dev->name);
	      dev->n2--;
	      if (dev->n2 == 0) {
		wlog(L_WARNING, "No response at device '%s'", dev->name);
		dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
		dev->status |= T_CALLEND;
		dev->offset = dev->pos_line = dev->pos_block = 0;
		dev->n2 = dev->i_n2;
	      } else
		if (pack = dev->head.stqh_first) {
		  pack->status &= ~P_WAITACK;
		  pack->status |= P_SENDING;
		  pack->offset = 0;
		  dev->status &= ~T_ACKWAIT;
		  dev->status |= T_BLOCKWAIT;
		}
	      dev->t3 = dev->i_t3;
	    }
	  }
	  if (!(dev->status & T_CALLOUT) && (dev->status & T_BLOCKWAIT)
	      && (dev->status & T_CONNECTED) && !(dev->status & T_IDENTWAIT)) {
	    /* Wait for new block */
	    dev->t4--;
	    if (dev->t4 == 0) {
	      wlog(L_DEBUG, "Next block wait timeout at device '%s'",
		   dev->name);
	      dev->n2--;
	      if (dev->n2 == 0) {
		wlog(L_WARNING, "No new blocks at device '%s'",
		     dev->name);
		dev->status &= ~(T_ACKWAIT | T_CRWAIT | T_BLOCKWAIT);
		dev->status |= T_CALLEND;
		dev->offset = dev->pos_line = dev->pos_block = 0;
		dev->n2 = dev->i_n2;
	      } else
		queue_pack_ack(dev);
	      dev->t4 = dev->i_t4;
	    }
	  }
	}
      }
    }
  } while (!flag_fatal_error && !flag_reread);
  wlog(L_INFO, "Exiting from main loop");
} /* loop */

/* Route packet (data) with length (len) to pager (dest) */
int route_pack(char *dest, unsigned char *data, int len, char p_func) {
  char filename[256];
  int fd;
  FILE *f;
  char filter[512];
  unsigned char pbuf[MAX_PACKET_SIZE];
  char buf[128];
  char *destroute;
  char *pdest;
  char *pdiv;
  char *paddr;
  tap_device *dev;
  tap_packet *pack;
  wo_packet *wpack;
  int pid, flag;
  int offset;
  struct stat fst;
  unsigned char *pdata;
  int plen;
  char pagerid[11];
  int error;
  int max_r_len;
  int slen;

  error = 0;
  pdata = strchr(data, CR);
  if (!pdata) {
    wlog(L_WARNING, "Packet format error or empty packet");
    return -1;
  }
  if (dest) {
    for (pdest = dest; *pdest && (*pdest <= ' '); pdest++);
    strncpy(pagerid, pdest, sizeof(pagerid));
    pagerid[sizeof(pagerid) - 1] = '\0';
    for (pdest = pagerid; *pdest && (*pdest >= ' '); pdest++);
    *pdest = '\0';
    /* Let's get destination from SQL-database */
    destroute = get_dest(pagerid);
  } else {
    destroute = NULL;
    if ((pdata - data) < sizeof(pagerid)) {
      memcpy(pagerid, data, pdata - data);
      pagerid[pdata - data] = '\0';
    } else {
      memcpy(pagerid, data, sizeof(pagerid) - 1);
      pagerid[sizeof(pagerid) - 1] = '\0';
    }
    dest = pagerid;
  }
  pdata++;
  plen = len - (pdata - data);
  if (!destroute) {
    wlog(L_WARNING, "Can't get destination for address '%s'",
	 dest);
    if (!defaultroute)
      return -1;
    destroute = malloc(strlen(defaultroute->name) + 5);
    if (!destroute)
      return -1;
    sprintf(destroute, "TAP:%s", defaultroute->name);
    error = -1;
  }
  pdest = destroute - 1;
  do {
    pdest++;
    paddr = strchr(pdest, ':');
    pdiv = strchr(pdest, ',');
    if (pdiv) {
      if (paddr && ((paddr + 1) < pdiv)) {
	if ((pdiv - paddr - 1) >= sizeof(buf))
	  pdiv = paddr + sizeof(buf);
	memcpy(buf, paddr + 1, pdiv - paddr - 1);
	buf[pdiv - paddr - 1] = '\0';
      } else
	buf[0] = '\0';
    } else {
      if (paddr) {
	strncpy(buf, paddr + 1, sizeof(buf));
	buf[sizeof(buf) - 1] = '\0';
      } else
	buf[0] = '\0';
    }
    if (!strncasecmp(pdest, "PAGER:", 6)) {
      char newid[11];
      char *pcin;
      char *pcout;
      char *newpcin;
      int first, last;

      /* Rewrite destination */
      pcout = newid;
      for (pcin = buf; *pcin; pcin++) {
	switch (*pcin) {
	case '{':
	  first = 0;
	  last = 0;
	  pcin++;
	  if (*pcin && ((*pcin <= '9') && (*pcin >= '0'))) {
	    first = strtol(pcin, &newpcin, 10) - 1;
	    pcin = newpcin;
	    if (*pcin) {
	      if (*pcin != '}') {
		if (*pcin == ';') {
		  pcin++;
		  if (*pcin && ((*pcin <= '9') && (*pcin >= '0'))) {
		    last = strtol(pcin, &newpcin, 10) - 1;
		    pcin = newpcin;
		  }
		  if (*pcin && (*pcin != '}'))
		    pcin--;
		}
	      }
	    }
	  }
	  if ((first < strlen(pagerid)) && (last < strlen(pagerid))) {
	    if (!last)
	      last = strlen(pagerid) - 1;
	    if (last < first)
	      last = first;
	    if ((last - first + 2 + (pcout - newid)) >= sizeof(newid))
	      last = sizeof(newid) - (pcout - newid) - 2 + first;
	    memcpy(pcout, pagerid + first, last - first + 1);
	    pcout = pcout + (last - first + 1);
	  } else
	    wlog(L_WARNING, "Wrong substitution parameter (in '%s') for '%s'",
		 buf, dest);
	  break;
	default:
	  *pcout = *pcin;
	  pcout++;
	}
	if (pcout >= (newid + sizeof(newid) - 1))
	  break;
	if (!*pcin)
	  break;
      }
      *pcout = '\0';
      /* Recursive route */
      if (!strcmp(pagerid, newid)) {
	wlog(L_WARNING, "Loop in route for '%s'! Packet dropped", dest);
	log_msg(data, len, "loop in route");
      } else {
	unsigned char *ptemp;
	int templen;

	ptemp = (unsigned char *) malloc(strlen(newid) + plen + 1);
	if (!ptemp)
	  wlog(L_WARNING, "Could not allocate memory for packet for '%s'",
	       dest);
	else {
	  wlog(L_DEBUG, "Packet for '%s' readressed to '%s'", dest, newid);
	  memcpy(ptemp + strlen(newid), pdata - 1, len - (pdata - data - 1));
	  memcpy(ptemp, newid, strlen(newid));
	  templen = plen + strlen(newid) + 1;
	  route_pack(newid, ptemp, templen, p_func);
	  free(ptemp);
	}
      }
    } else
      if (!strncasecmp(pdest, "TAP:", 4)) {
	/* Route to one of rtapd's devices */
	max_r_len = 0;
	dev = check_avail_dev(buf, &max_r_len);
	if (dev) {
	  if (dev->npacks >= maxpacks)
	    wlog(L_WARNING, "Queue full (too many packets) at device '%s'",
		 dev->name);
	  else {
	    wlog(L_INFO, "Packet for '%s' routed to device '%s'",
		 dest, dev->name);
	    if (dev->type == 0) {
	      unsigned char *pmsg;

	      if ((max_r_len > 0) && (max_r_len < len)) {
		wlog(L_INFO, "Truncating message to length %d", max_r_len);
		slen = max_r_len;
	      } else
		slen = len;
	      pmsg = malloc(slen);
	      if (pmsg)
	        memcpy(pmsg, data, slen);
	      /* To serial device */
	      for (offset = 0; offset < slen;) {
		pack = get_next_block(data, &offset, slen);
		if (pack) {
		  pack->nc = dev->i_nc;
		  pack->msg = pmsg;
		  pack->msglen = slen;
		  if (dev->crlf || dev->lfcr)
		    conv_cr(pack, dev->crlf);
		  STAILQ_INSERT_TAIL(&dev->head, pack, packp);
		  dev->npacks++;
		}
	      }
	    } else {
	      /* To file or external filter. First we must fork */
	      pid = fork();
	      if (pid == 0) {
		/* Write packet in child process (to prevent delays) */
		unsigned char *edata;
		int elen;
		char *pc1;
		char *pc2;
		int count;

		/* unescape characters */
		edata = (char *) malloc(plen + 1);
		if (!edata) {
		  edata = pdata;
		  elen = plen;
		} else {
		  pc1 = pdata;
		  pc2 = edata;
		  count = 0;
		  elen = 0;
		  while (count < plen) {
		    if (*pc1 == '\032') {
		      pc1++;
		      count++;
		      *pc2 = *pc1 - 0x40;
		    } else
		      *pc2 = *pc1;
		    elen++;
		    pc2++;
		    pc1++;
		    count++;
		  }
		}
		
		if (edata[elen - 1] == CR)
		  elen--;
		if (dev->type == 1) {
		  /* To file */
		  check_name(dest);
		  flag = 1;
		  if ((stat(dev->device, &fst) >= 0)
		      && (fst.st_mode & S_IFDIR)) {
		    flag = 0;
		    snprintf(filename, sizeof(filename), "%s/%s",
			     dev->device, dest);
		  } else
		    snprintf(filename, sizeof(filename), "%s", dev->device);
		  wlog(L_INFO, "Packet to '%s' saved in file '%s'",
		       dest, filename);
		  if ((fd = open(filename, O_WRONLY | O_CREAT
				 | O_APPEND, 0644)) >= 0) {
		    if ((flag && ((write(fd, dest,
					 strlen(dest)) != strlen(dest))
				  || (write(fd, "\n", 1) != 1)))
			|| (write(fd, edata, elen) != elen)
			|| (write(fd, "\n", 1) != 1))
		      wlog(L_WARNING, "Write to file '%s' error %s (%d)",
			   filename, strerror(errno), errno);
		    close(fd);
		  } else
		    wlog(L_WARNING, "Open file '%s' error %s (%d)",
			 filename, strerror(errno), errno);
		} else
		  if (dev->type == 2) {
		    /* To external filter */
		    paddr = strchr(buf, ':');
		    if (!paddr)
		      paddr = "";
		    check_name(dest);
		    snprintf(filter, sizeof(filter) - 1, "%s '%s' '%s'",
			     dev->device, dest, paddr + 1);
		    f = popen(filter, "w");
		    if (f) {
		      fd = fileno(f);
		      if (write(fd, edata, elen) != elen)
			wlog(L_WARNING, "Write to filter '%s' error %s (%d)",
			     filter, strerror(errno), errno);
		      pclose(f);
		    } else
		      wlog(L_WARNING, "Open pipe to '%s' error %s (%d)",
			   filter, strerror(errno), errno);
		  }
		if (edata != pdata)
		  free(edata);
		exit(0);
	      } else
		if (pid < 0)
		  wlog(L_WARNING, "Fork error %s (%d)",
		       strerror(errno), errno);
	    }
	  }
	} else {
	  /* No free device for this packet. Put it in wo queue */
	  if (nwpacks >= max_wopacks)
	    wlog(L_WARNING, "Queue for delayed packets is full!");
	  else {
	    unsigned char *pmsg;

	    if ((max_r_len > 0) && (max_r_len < len)) {
	      wlog(L_INFO, "Truncating message to length %d", max_r_len);
	      slen = max_r_len;
	    } else
	      slen = len;
	    pmsg = malloc(slen);
	    if (pmsg)
	      memcpy(pmsg, data, slen);
	    wlog(L_DEBUG, "Packet for '%s' delayed", dest);
	    offset = 0;
	    for (offset = 0; offset < slen;) {
	      pack = get_next_block(data, &offset, slen);
	      if (pack) {
		pack->msg = pmsg;
		pack->msglen = slen;
		wpack = (wo_packet *) malloc(sizeof(wo_packet));
		if (wpack) {
		  wpack->pack = pack;
		  strncpy(wpack->device, buf, sizeof(wpack->device));
		  STAILQ_INSERT_TAIL(&wpacks, wpack, packp);
		  nwpacks++;
		}
	      }
	    }
	  }
	}
      } else
	if (!strncasecmp(pdest, "TNPP:", 5)) {
	  /* Route to rtnppd */
	  wlog(L_INFO, "Packet for '%s' to TNPP '%s'", dest, buf);
	  pid = fork();
	  if (pid == 0) {
	    if (p_func == sheduler_func)
	      sprintf(pbuf, "B%c%- 10s", page_func, pagerid);
	    else
	      sprintf(pbuf, "B%c%- 10s", p_func, pagerid);
	    memcpy(pbuf + 12, pdata, plen);
	    pbuf[plen + 12] = ETB;
	    if (send_tnpp_packet(tnpp_out, &u_sun, pbuf, plen + 13,
				 strtol(buf, NULL, 16), my_tnpp_addr) <= 0)
	      wlog(L_WARNING, "Send to unix socket error %s (%d)",
		   strerror(errno), errno);
	    exit(0);
	  } else
	    if (pid < 0)
	      wlog(L_WARNING, "Fork error %s (%d)", strerror(errno), errno);
	} else
	  if (!strncasecmp(pdest, "SHEDULER:", 9)) {
	    /* External sheduler */
	    dev = check_avail_dev(buf, NULL);
	    if ((p_func == sheduler_func) || !dev) {
	      wlog(L_INFO, "Packet for '%s' from sheduler. Next destination",
		   dest);
	    } else {
	      char *pparam;

	      paddr = strchr(buf, ':');
	      if (!paddr) {
		if (!use_file_table) {
		  char query[256];
		
		  snprintf(query, sizeof(query) - 1, "SELECT shedule FROM"
			   " shedules WHERE pager = '%s'", dest);
		  pparam = run_query(query);
		} else 
		  pparam = NULL;
	      } else
		pparam = strdup(paddr + 1);
	      if (!pparam || !strlen(pparam)) {
		wlog(L_INFO, "'%s' has no shedules. Next destination",
		     dest);
	      } else {
		/* To sheduler's filter. First we must fork */
		pid = fork();
		if (pid == 0) {
		  /* Write packet in child process (to prevent delays) */
		  if (data[len - 1] == CR)
		    len--;
		  check_name(dest);
		  snprintf(filter, sizeof(filter) - 1, "%s '%s' '%s'",
			   dev->device, dest, pparam);
		  f = popen(filter, "w");
		  if (f) {
		    fd = fileno(f);
		    if (write(fd, pdata, plen) != plen)
		      wlog(L_WARNING, "Write to filter '%s' error %s (%d)",
			   filter, strerror(errno), errno);
		    pclose(f);
		  } else
		    wlog(L_WARNING, "Open pipe to '%s' error %s (%d)",
			 filter, strerror(errno), errno);
		} else {
		  if (pid < 0)
		    wlog(L_WARNING, "Fork error %s (%d)",
			 strerror(errno), errno);
		  else {
		    wlog(L_INFO, "Packet for '%s' routed to sheduler", dest);
		    break;
		  }
		}
		exit(0);
	      }
	      if (pparam)
		free(pparam);
	    }
	  } else
	    if (!strncasecmp(pdest, "NONE", 4)) {
	      /* Junk packet */
	      wlog(L_INFO, "Packet for '%s' dropped", dest);
	    } else {
	      wlog(L_WARNING, "Unknown device's name in route for '%s'",
		   dest);
	      log_msg(data, len, "unknown device in route record");
	    }
    pdest = strchr(pdest, ',');
  } while(pdest);
  free(destroute);
  return error;
} /* route_pack */

/* Log string via syslog */
void wlog(int prio, char *fmt,...) {
  va_list ap;
  char buf[512];
  time_t ltime;
  char *timestamp;
  char buftime[16];

  if (loglevel >= prio) {
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    buf[sizeof(buf) - 1] = '\0';
    if (log_stderr) {
      time(&ltime);
      timestamp = ctime(&ltime);
      memcpy(buftime, timestamp + 4, 15);
      buftime[15] = '\0';
      fprintf(stderr, "%s %s\n", buftime, buf);
    } else
      syslog(prio, buf);
  }
} /* wlog */

void log_msg(unsigned char *msg, int len, char *reason) {
  FILE *f;
  unsigned char *p;
  unsigned char buf[MAX_PACKET_SIZE];
  time_t ltime;
  char *timestamp;
  char buftime[16];

  f = fopen(file_bad, "w");
  if (!f)
    wlog(L_WARNING, "Error '%s' (%d) at open file '%s'", strerror(errno),
	 errno, file_bad);
  else {
    p = strchr(msg, CR);
    if (p && ((p - msg) < sizeof(buf))) {
      time(&ltime);
      timestamp = ctime(&ltime);
      memcpy(buftime, timestamp + 4, 15);
      buftime[15] = '\0';
      memcpy(buf, msg, p - msg);
      buf[p - msg] = '\0';
      p++;
      fprintf(f, "%s %s: '%s'", buftime, reason, buf);
      memcpy(buf, p, len - (p - msg));
      buf[len - (p - msg)] = '\0';
      fprintf(f, " '%s'\n", buf);
    }
    fclose(f);
  }
  route_pack(NULL, msg, len, '@');
} /* log_msg */

/* Log data packets */
void log_data(char *type, unsigned char *data, int len) {
  int i;
  unsigned char buf1[384];
  unsigned char buf2[192];
  unsigned char hex[4];
  unsigned char c;

  if (!log_all_data)
    return;
  buf1[0] = '\0';
  buf2[0] = '\0';
  for (i = 0; i < len; i++) {
    c = data[i];
    if (strlen(buf1) < (sizeof(buf1) - 4)) {
      sprintf(hex, "%02X ", c);
      strcat(buf1, hex);
    }
    if (strlen(buf2) < (sizeof(buf2) - 2)) {
      if (c >= ' ')
	hex[0] = c;
      else
	hex[0] = '.';
      hex[1] = '\0';
      strcat(buf2, hex);
    }
  }
  wlog(L_DEBUG, "Data %s: %s'%s'", type, buf1, buf2);
} /* log_data */

/* Process new tap packet */
int process_tap_packet(unsigned char *data, int len) {
  char dest[11];
  int i;

  for (i = 0; (i < 10) && (data[i] != CR); i++)
    dest[i] = data[i];
  dest[i] = '\0';
  if (data[i] != CR) {
    wlog(L_WARNING, "Bad pager id number");
    return 0;
  }
  if (route_pack(dest, data, len, '@'))
    return 0;
  return 1;
} /* process_tap_packet */

/* (Re)open serial port device */
void reconnect(tap_device *dev) {
  int flag;
  tap_device *tdev;
  tap_device *master;

  if (dev->type != 0)
    return;
  if (dev->fd >= 0) {
    flag = fcntl(dev->fd, F_GETFL);
    fcntl(dev->fd, F_SETFL, flag | O_NONBLOCK);
    flag = 0;
    ioctl(dev->fd, TIOCFLUSH, &flag);
    close(dev->fd);
  }
  dev->fd = open(dev->device, O_RDWR | O_NONBLOCK);
  if (dev->fd < 0) {
    wlog(L_ERROR, "Open file '%s' error %s (%d) at device '%s'",
	 dev->device, strerror(errno), errno, dev->name);
    flag_fatal_error = 1;
    return;
  }
  if (dev->master) {
    master = dev->master;
    master->fd = dev->fd;
  } else
    master = dev;
  for (tdev = head.stqh_first; tdev;
       tdev = tdev->devp.stqe_next)
    if ((tdev->master == master) && (tdev != dev))
      tdev->fd = dev->fd;
  set_serial(dev);
  flag = TIOCM_DTR;
  if ((dev->init1 && dev->init1[0]) || !dev->dtrtoggle)
    ioctl(dev->fd, TIOCMBIS, &flag);
  else
    ioctl(dev->fd, TIOCMBIC, &flag);
  if (dev->clocal)
    dev->status |= T_CONNECTED;
} /* reconnect */

/* Set parity on serial port */
void set_port_parity(int fd, int bits, int parity) {
  struct termios term;

  if (fd < 0)
    return;
  tcgetattr(fd, &term);
  term.c_cflag &= ~(CSIZE | PARENB | PARODD);
  term.c_iflag &= ~(ISTRIP | IGNPAR);
  if (!parity) {
    term.c_cflag |= CS8;
    if (bits == 7)
      term.c_iflag |= ISTRIP;
  } else {
    term.c_cflag |= (CS7 | PARENB);
    term.c_iflag |= IGNPAR;
  }
  tcsetattr(fd, TCSANOW, &term);
} /* set_port_parity */

/* Set serial parameters */
void set_serial(tap_device *dev) {
  int res;
  struct termios term;
  const struct {
    int baud;
    speed_t t_baud;
  } speeds[12] = {{300, B300}, {600, B600}, {1200, B1200}, {2400, B2400},
		  {4800, B4800}, {7200, B7200}, {9600, B9600}, {19200, B19200},
		  {38400, B38400}, {57600, B57600}, {115200, B115200},
		  {9600, B9600}};
  int i;

  if (dev->type != 0)
    return;
  if (dev->fd < 0) {
    /* Hmmm... Not opened */
    reconnect(dev);
    return;
  }
  tcgetattr(dev->fd, &term);
  for (i = 0; i < 11; i++)
    if (speeds[i].baud == dev->baud)
      break;
  cfsetispeed(&term, speeds[i].t_baud);
  cfsetospeed(&term, speeds[i].t_baud);
  term.c_cflag &= ~(CSTOPB | PARENB | PARODD | HUPCL | CLOCAL | CRTSCTS
		    | CSIZE);
  term.c_iflag = 0;
  term.c_oflag = 0;
  term.c_lflag = 0;
  term.c_cflag |= CREAD | CLOCAL;
  if (dev->parity) {
    term.c_cflag |= PARENB | CS7;
    term.c_iflag |= IGNPAR;
  } else {
    term.c_cflag |= CS8;
    term.c_iflag |= ISTRIP;
  }
  if (dev->xon_xoff)
    term.c_iflag |= IXON | IXOFF;
  if (dev->crtscts)
    term.c_cflag |= CRTSCTS;
  term.c_cc[VMIN] = 1;
  term.c_cc[VTIME] = 0;
  if (tcsetattr(dev->fd, TCSANOW, &term) < 0) {
    wlog(L_ERROR, "Set serial parameters error %s (%d) at device '%s'",
	 strerror(errno), errno, dev->name);
    flag_fatal_error = 1;
  }
  dev->tr = dev->i_tr;
} /* set_serial */

/* Return destination for pager-id (in malloc'ed buffer) */
char *get_dest(char *dest) {
  char *p;

  if (use_file_table) {
    /* Look for routes in previously readed file */
    fixed_route *route;

    for (route = routes.stqh_first; route; route = route->next.stqe_next)
      if ((route->from == my_tnpp_addr) || (route->from == 0)) {
	if (!strcmp(route->to, dest))
	    goto FOUND_FIXED;
	p = strchr(route->to, '-');
	if (p) {
	  char buf[22];

	  memcpy(buf, route->to, p - route->to);
	  buf[p - route->to] = '\0';
	  if ((strcmp(dest, buf) >= 0) && (strcmp(dest, p + 1) <= 0))
	    goto FOUND_FIXED;
	}
      }
    return NULL;
  FOUND_FIXED:
    p = strdup(route->destination);
    return p;
  } else {
    /* Look for destination in SQL-database */
    char query[256];

    snprintf(query, sizeof(query) - 1, "select xto, xfrom, xwhere, xwhereh"
	     " from %s where"
	     " (xwhere = '%s' or (xwhere < '%s' and xwhereh >= '%s'))"
	     " and (xfrom = %d or xfrom = 0)"
	     " order by xfrom desc, xwhereh asc, xwhere desc",
	     "routes", dest, dest, dest, my_tnpp_addr);
    return (run_query(query));
  }
  return NULL;
} /* get_dest */

/* Get data from SQL-database */
char *run_query(char *query) {
#if defined(USE_MSQL) || defined(USE_POSTGRESQL) || defined(USE_MYSQL)
  int res;
  char *ans;
  char *s;
#ifdef USE_MSQL
  m_result *qr;
  m_row row;
#else
#ifdef USE_POSTGRESQL
  PGresult *pgr;
#else
#ifdef USE_MYSQL
  MYSQL_RES *mqr;
  MYSQL_ROW row;
#endif
#endif
#endif

  while
#ifdef USE_MSQL
	 ((res = msqlQuery(sock_sql, query)) < 0)
#else
#ifdef USE_POSTGRESQL
	 (!(pgr = PQexec(conn_sql, query)))
#else
#ifdef USE_MYSQL
	 (mysql_query(conn_sql, query))
#endif
#endif
#endif
    {
      wlog(L_WARNING, "Trying reconnect to SQL-server");
      to_sql_disconnect();
      if (to_sql_connect()) {
	flag_fatal_error = 1;
	return NULL;
      }
    }
#ifdef USE_MSQL
  if (res == 0)
    return NULL;
  qr = msqlStoreResult();
  if (!qr) {
    wlog(L_WARNING, "Can't allocate memory for result");
    return NULL;
  }
  row = msqlFetchRow(qr);
  if (!row || !row[0])
    return NULL;
  ans = strdup(row[0]);
  msqlFreeResult(qr);
#else
#ifdef USE_POSTGRESQL
  if ((PQresultStatus(pgr) != PGRES_TUPLES_OK)
      || !(s = PQgetvalue(pgr, 0, 0)))
    return NULL;
  ans = strdup(s);
  PQclear(pgr);
#else
#ifdef USE_MYSQL
  mqr = mysql_store_result(conn_sql);
  if (!mqr) {
    wlog(L_WARNING, "Can't allocate memory for result");
    return NULL;
  }
  row = mysql_fetch_row(mqr);
  if (!row || !row[0])
    return NULL;
  ans = strdup(row[0]);
  mysql_free_result(mqr);
#endif
#endif
#endif
  if (ans)
    for (s = ans + strlen(ans) - 1; (s != ans) && (*s == ' '); s--)
      *s = '\0';
  return ans;
#endif
} /* run_query */


/* rtapd startup routines */

int to_sql_connect() {
  int res;

#ifdef USE_MSQL
  sock_sql = msqlConnect(sql_serv);
  if (sock_sql < 0) {
    wlog(L_ERROR, "Can't connect to SQL-server (%s)", msqlErrMsg);
    return -1;
  } else {
    res = msqlSelectDB(sock_sql, sql_base);
    if (res < 0) {
      wlog(L_ERROR, "Can't select database %s (%s)", sql_base, msqlErrMsg);
      return -1;
    }
  }
#else
#ifdef USE_POSTGRESQL
  conn_sql = PQsetdbLogin(sql_serv, sql_port, NULL, NULL,
			  sql_base, sql_user, sql_passwd);
  if (!conn_sql) {
    wlog(L_ERROR, "Unknown error at connect to SQL-server");
    return -1;
  } else
    if (PQstatus(conn_sql) == CONNECTION_BAD) {
      wlog(L_ERROR, "Can't connect ot SQL-server (%s)",
	   PQerrorMessage(conn_sql));
      return -1;
    }
#else
#ifdef USE_MYSQL
  conn_sql = mysql_init(conn_sql);
  conn_sql = mysql_real_connect(conn_sql, sql_serv, sql_user, sql_passwd,
				sql_base, atoi(sql_port), NULL, 0);
  if (!conn_sql) {
    wlog(L_ERROR, "Unknown error at connect to SQL-server");
    return -1;
  }
#endif
#endif
#endif
  use_sql = 1;
  return 0;
} /* to_sql_connect */

tap_device defaults;    /* Defaults for all devices */
tap_device curr_dev;

/* Free all memory allocated for device (with packets queue) */
void destroy_device(tap_device *dev) {
  tap_packet *pack;
  unsigned char *p;
  int flag;

  if (!dev)
    return;
  if (dev != &defaults)
    wlog(L_DEBUG, "Clearing memory for device '%s'", dev->name);
  if (dev->fd >= 0) {
    if (dev->dtrtoggle)
      DTR_down(dev->fd);
    uucp_unlock(dev->device);
    flag = fcntl(dev->fd, F_GETFL);
    fcntl(dev->fd, F_SETFL, flag | O_NONBLOCK);
    flag = 0;
    ioctl(dev->fd, TIOCFLUSH, &flag);
    close(dev->fd);
  }
  /* Free packets */
  p = NULL;
  while (dev->head.stqh_first) {
    pack = dev->head.stqh_first;
    STAILQ_REMOVE_HEAD(&dev->head, packp);
    if (pack->msg && pack->msg != p) {
      free(pack->msg);
      p = pack->msg;
    }
    free(pack);
  }
  /* Free private device's data */
  if ((dev->device != dev->name) && (dev != &defaults))
    free(dev->name);
  if (dev->device && (dev != &defaults))
    free(dev->device);
  if (dev->init1)
    free(dev->init1);
  if (dev->init2)
    free(dev->init2);
  if (dev->dial_p)
    free(dev->dial_p);
  if (dev->dial_s)
    free(dev->dial_s);
  if (dev->passwd)
    free(dev->passwd);
  if (dev != &defaults)
    free(dev);
} /* destroy_device */

/* Destroy all devices in stack */
void destroy_all_devices() {
  tap_device *dev;
  wo_packet *wpack;

  while (head.stqh_first) {
    dev = head.stqh_first;
    STAILQ_REMOVE_HEAD(&head, devp);
    destroy_device(dev);
  }
  STAILQ_INIT(&head);
  defaults.fd = -1;
  destroy_device(&defaults);
  memset(&defaults, 0, sizeof(tap_device));
  while (wpacks.stqh_first) {
    wpack = wpacks.stqh_first;
    STAILQ_REMOVE_HEAD(&wpacks, packp);
    free(wpack->pack);
    free(wpack);
  }
  STAILQ_INIT(&wpacks);
  nwpacks = 0;
} /* destory_all_device */

/* Print string value */
void dump_string(FILE *out, char *name, char *value) {
  char buf[256];
  char *pc;
  char *pco;

  pc = value;
  if (value) {
    for (pco = buf; (*pc) && (pco < (buf + sizeof(buf))); pc++)
      if ((*pc < ' ') || (*pc == '\\')) {
	*pco++ = '\\';
	switch (*pc) {
	case '\r':
	  *pco++ = 'r';
	  break;
	case '\t':
	  *pco++ = 't';
	  break;
	case '\n':
	  *pco++ = 'n';
	  break;
	case '\\':
	default:
	  *pco++ = '\\';
	}
      } else
	*pco++ = *pc;
  } else
    pco = buf;
  *pco = '\0';
  fprintf(out, "%s '%s'\n", name, buf);
} /* dump_string */

/* Print boolean value */
void dump_bool(FILE *out, char *name, int value) {
  fprintf(out, "%s %s\n", name, value ? "On" : "Off");
} /* dump_bool */

/* Dump device configuration to file */
void dump_device_conf(FILE *out, tap_device *dev) {
  fprintf(out, "Device %s\n", dev->device);
  fprintf(out, "DeviceName %s\n", dev->name);
  if (!dev->type) {
    dump_bool(out, "Dial", dev->dial);
    dump_bool(out, "CLocal", dev->clocal);
    dump_string(out, "Init1", dev->init1);
    dump_string(out, "Init2", dev->init2);
    dump_string(out, "DialPrefix", dev->dial_p);
    dump_string(out, "DialSuffix", dev->dial_s);
    if (dev->passwd)
      fprintf(out, "Password %s\n", dev->passwd);
    fprintf(out, "Baud %d\n", dev->baud);
    dump_bool(out, "CRtsCts", dev->crtscts);
    dump_bool(out, "XonXoff", dev->xon_xoff);
    dump_bool(out, "Parity", dev->parity);
    dump_bool(out, "AutoDetect", dev->autodetect);
    dump_bool(out, "DTRToggle", dev->dtrtoggle);
    dump_bool(out, "CRLF", dev->crlf);
    dump_bool(out, "LFCR", dev->lfcr);
    dump_bool(out, "NoIncoming", dev->no_incoming);
    fprintf(out, "MaxBlocks %d\n", dev->max_blocks);
    fprintf(out, "MaxTransactions %d\n", dev->max_trans);
    fprintf(out, "CallRetry %d\n", dev->i_nc);
    fprintf(out, "NRetry %d\n", dev->i_nr);
    fprintf(out, "BlockTime %d\n", dev->i_tblock);
    fprintf(out, "SessionTime %d\n", dev->i_tsess);
    fprintf(out, "TimeOut %d\n", dev->i_tr);
    fprintf(out, "T1 %d\n", dev->i_t1);
    fprintf(out, "T2 %d\n", dev->i_t2);
    fprintf(out, "T3 %d\n", dev->i_t3);
    fprintf(out, "T4 %d\n", dev->i_t4);
    fprintf(out, "T5 %d\n", dev->i_t5);
    fprintf(out, "N1 %d\n", dev->i_n1);
    fprintf(out, "N2 %d\n", dev->i_n2);
    fprintf(out, "N3 %d\n", dev->i_n3);
  }
} /* dump_device_conf */

/* Dump main configuration */
void dump_main_conf(FILE *out) {
  fprintf(out, "TNPPAddress %04X\n", my_tnpp_addr);
  fprintf(out, "ExtSocketIn %s\n", u_sock_name);
  fprintf(out, "ExtSocketOut %s\n", u_sock_name_out);
  fprintf(out, "MaxQueueSize %d\n", maxpacks);
  fprintf(out, "MaxQueueDelayed %d\n", max_wopacks);
  fprintf(out, "LockDir %s\n", lock_dir);
  fprintf(out, "DeviceFile %s\n", file_dev_conf);
  fprintf(out, "RunCommand %s\n", command_filter);
  fprintf(out, "BadMessages %s\n", file_bad);
  fprintf(out, "AllowSendAll %s\n", allowsend ? "Yes" : "No");
  fprintf(out, "ShedulerFunc %02X\n", sheduler_func);
  if (file_route_table)
    fprintf(out, "RoutingTable %s\n", file_route_table);
  else {
    fprintf(out, "PidFile %s\n", file_pid);
    if (sql_serv)
      fprintf(out, "SQLServer %s\n", sql_serv);
    fprintf(out, "SQLPort %s\n", sql_port);
    fprintf(out, "SQLDataBase %s\n", sql_base);
    if (sql_user)
      fprintf(out, "SQLUser %s\n", sql_user);
    if (sql_passwd)
      fprintf(out, "SQLPassword %s\n", sql_passwd);
  }
} /* dump_main_conf */  

/* Dump route record */
void dump_route_conf(FILE *out, fixed_route *route) {
  fprintf(out, "%04X %s %s\n", route->from, route->to, route->destination);
} /* dump_route_conf */

/* Dump all configuration data in file */
void dump_config(char *filename) {
  FILE *out;
  tap_device *dev;
  fixed_route *route;

  out = fopen(filename, "w");
  if (!out) {
    wlog(L_ERROR, "Can't open file '%s' to dump configuration", filename);
    return;
  }
  fprintf(out, "; rtapd.conf\n\n");
  dump_main_conf(out);
  fprintf(out, "\n\n; tap_dev.conf\n\n");
  for (dev = head.stqh_first; dev; dev = dev->devp.stqe_next) {
    dump_device_conf(out, dev);
    fprintf(out, "\n");
  }
  fprintf(out, "\n; tap_route.conf\n\n");
  if (file_route_table)
    for (route = routes.stqh_first; route; route = route->next.stqe_next)
      dump_route_conf(out, route);
  else
    fprintf(out, "; No static routing data (look in SQL database)\n");
  fclose(out);
} /* dump_config */

/* Allocate memory for new device */
tap_device *create_device() {
  tap_device *dev;

  dev = (tap_device *) malloc(sizeof(tap_device));
  return dev;
} /* create_device */

int flag_default;

/* Copy device's config into device's stack */
void copy_curr_dev() {
  tap_device *dev;
  tap_device *tdev;

  if (curr_dev.device) {
    if (!curr_dev.device[0]) {
      /* Change default settings */
      memcpy(&defaults, &curr_dev, sizeof(tap_device));
    } else {
      /* Save device's settings */
      dev = create_device();
      if (!dev)
	goto MALLOC_ERROR;
      memcpy(dev, &curr_dev, sizeof(tap_device));
      STAILQ_INSERT_TAIL(&head, dev, devp);
      STAILQ_INIT(&dev->head);
      dev->master = NULL;
      for (tdev = head.stqh_first; tdev != dev;
	   tdev = tdev->devp.stqe_next)
	if (!strcmp(tdev->device, dev->device)) {
	  dev->master = tdev;
	  break;
	}
      memcpy(&curr_dev, &defaults, sizeof(tap_device));
      if (flag_default) {
	defaultroute = dev;
	flag_default = 0;
      }
    }
    /* Dup some settings */
    if (defaults.init1) {
      curr_dev.init1 = strdup(defaults.init1);
      if (!curr_dev.init1)
	goto MALLOC_ERROR;
    }
    if (defaults.init2) {
      curr_dev.init2 = strdup(defaults.init2);
      if (!curr_dev.init2)
	goto MALLOC_ERROR;
    }
    if (defaults.dial_p) {
      curr_dev.dial_p = strdup(defaults.dial_p);
      if (!curr_dev.dial_p)
	goto MALLOC_ERROR;
    }
    if (defaults.dial_s) {
      curr_dev.dial_s = strdup(defaults.dial_s);
      if (!curr_dev.dial_s)
	goto MALLOC_ERROR;
    }
    if (defaults.passwd) {
      curr_dev.passwd = strdup(defaults.passwd);
      if (!curr_dev.passwd)
	goto MALLOC_ERROR;
    }
  } else {
    /* Defaults parameters */
    curr_dev.device = "";
    curr_dev.name = NULL;
    curr_dev.type = 0;
    curr_dev.status = 0;
    curr_dev.phone[0] = '\0';
    curr_dev.passwd = NULL;
    curr_dev.offset = curr_dev.pos_line = curr_dev.pos_block = 0;
    STAILQ_INIT(&curr_dev.head);
    curr_dev.npacks = 0;
    curr_dev.devp.stqe_next = NULL;
    curr_dev.lastack = 0;
    curr_dev.lastdial = 0;
    curr_dev.next_multi = 0;
    curr_dev.crlf = 0;
    curr_dev.lfcr = 0;
    curr_dev.fd = -1;
    curr_dev.init1 = curr_dev.init2 = curr_dev.dial_p = curr_dev.dial_s = NULL;
    curr_dev.i_t1 = curr_dev.t1 = 2;
    curr_dev.i_t2 = curr_dev.t2 = 1;
    curr_dev.i_t3 = curr_dev.t3 = 10;
    curr_dev.i_t4 = curr_dev.t4 = 4;
    curr_dev.i_t5 = curr_dev.t5 = 8;
    curr_dev.i_n1 = curr_dev.n1 = 3;
    curr_dev.i_n2 = curr_dev.n2 = 3;
    curr_dev.i_n3 = curr_dev.n3 = 3;
    curr_dev.i_tblock = 120;
    curr_dev.i_tsess = 600;
    curr_dev.tblock = 0;
    curr_dev.i_tr = curr_dev.tr = 60;
    curr_dev.i_nr = curr_dev.nr = 16;
    curr_dev.i_nc = 8;
    curr_dev.max_blocks = 16;
    curr_dev.max_trans = 8;
    curr_dev.clocal = 0;
    curr_dev.dial = 0;
    curr_dev.parity = 1;
    curr_dev.autodetect = 0;
    curr_dev.baud = 300;
    curr_dev.xon_xoff = 1;
    curr_dev.crtscts = 0;
    curr_dev.dtrtoggle = 1;
    curr_dev.no_incoming = 0;
    curr_dev.master = NULL;
    return;
  }
  return;
 MALLOC_ERROR:
  wlog(L_ERROR, "Malloc failed at allocating memory for new device");
  exit(2);
} /* copy_curr_dev */

int add_badroute(void *addr, int len, int flag, void *value) {
  flag_default = 1;
  return 1;
} /* add_badroute */

int add_device(void *addr, int len, int flag, void *value);
#define DEVICE_CONF_NUM 34
conf_rec devices_conf[DEVICE_CONF_NUM] = {
  {PARAM("device"), STRING, (void *) NULL,
   0, NULL, 0, add_device},
  {PARAM("devicename"), STRING_NC, (void *) &curr_dev.name,
   0, NULL, 0, std_STRING},
  {PARAM("baud"), DEC_DIGITS, (void *) &curr_dev.baud,
   0, NULL, 0, std_DIGITS},
  {PARAM("clocal"), BOOL_PAR, (void *) &curr_dev.clocal,
   0, NULL, 0, std_BOOL},
  {PARAM("dtrtoggle"), BOOL_PAR, (void *) &curr_dev.dtrtoggle,
   0, NULL, 0, std_BOOL},
  {PARAM("crtscts"), BOOL_PAR, (void *) &curr_dev.crtscts,
   0, NULL, 0, std_BOOL},
  {PARAM("xonxoff"), BOOL_PAR, (void *) &curr_dev.xon_xoff,
   0, NULL, 0, std_BOOL},
  {PARAM("parity"), BOOL_PAR, (void *) &curr_dev.parity,
   0, NULL, 0, std_BOOL},
  {PARAM("autodetect"), BOOL_PAR, (void *) &curr_dev.autodetect,
   0, NULL, 0, std_BOOL},
  {PARAM("dial"), BOOL_PAR, (void *) &curr_dev.dial,
   0, NULL, 0, std_BOOL},
  {PARAM("init1"), STRING_WQ, (void *) &curr_dev.init1,
   0, NULL, 0, std_STRING},
  {PARAM("init2"), STRING_WQ, (void *) &curr_dev.init2,
   0, NULL, 0, std_STRING},
  {PARAM("dialprefix"), STRING_WQ, (void *) &curr_dev.dial_p,
   0, NULL, 0, std_STRING},
  {PARAM("dialsuffix"), STRING_WQ, (void *) &curr_dev.dial_s,
   0, NULL, 0, std_STRING},
  {PARAM("password"), STRING, (void *) &curr_dev.passwd,
   0, NULL, 0, std_STRING},
  {PARAM("maxblocks"), DEC_DIGITS, (void *) &curr_dev.max_blocks,
   0, NULL, 0, std_DIGITS},
  {PARAM("maxtransactions"), DEC_DIGITS, (void *) &curr_dev.max_trans,
   0, NULL, 0, std_DIGITS},
  {PARAM("crlf"), BOOL_PAR, (void *) &curr_dev.crlf,
   0, NULL, 0, std_BOOL},
  {PARAM("lfcr"), BOOL_PAR, (void *) &curr_dev.lfcr,
   0, NULL, 0, std_BOOL},
  {PARAM("t1"), DEC_DIGITS, (void *) &curr_dev.i_t1,
   0, NULL, 0, std_DIGITS},
  {PARAM("t2"), DEC_DIGITS, (void *) &curr_dev.i_t2,
   0, NULL, 0, std_DIGITS},
  {PARAM("t3"), DEC_DIGITS, (void *) &curr_dev.i_t3,
   0, NULL, 0, std_DIGITS},
  {PARAM("t4"), DEC_DIGITS, (void *) &curr_dev.i_t4,
   0, NULL, 0, std_DIGITS},
  {PARAM("t5"), DEC_DIGITS, (void *) &curr_dev.i_t5,
   0, NULL, 0, std_DIGITS},
  {PARAM("n1"), DEC_DIGITS, (void *) &curr_dev.i_n1,
   0, NULL, 0, std_DIGITS},
  {PARAM("n2"), DEC_DIGITS, (void *) &curr_dev.i_n2,
   0, NULL, 0, std_DIGITS},
  {PARAM("n3"), DEC_DIGITS, (void *) &curr_dev.i_n3,
   0, NULL, 0, std_DIGITS},
  {PARAM("nretry"), DEC_DIGITS, (void *) &curr_dev.i_nr,
   0, NULL, 0, std_DIGITS},
  {PARAM("callretry"), DEC_DIGITS, (void *) &curr_dev.i_nc,
   0, NULL, 0, std_DIGITS},
  {PARAM("blocktime"), DEC_DIGITS, (void *) &curr_dev.i_tblock,
   0, NULL, 0, std_DIGITS},
  {PARAM("sessiontime"), DEC_DIGITS, (void *) &curr_dev.i_tsess,
   0, NULL, 0, std_DIGITS},
  {PARAM_NOP("badroute"), BOOL_NOP, NULL,
   0, NULL, 0, add_badroute},
  {PARAM("timeout"), DEC_DIGITS, (void *) &curr_dev.i_tr,
   0, NULL, 0, std_DIGITS},
  {PARAM("noincoming"), BOOL_PAR, (void *) &curr_dev.no_incoming,
   0, NULL, 0, std_BOOL}
};

int add_device(void *addr, int len, int flag, void *value) {
  char *s;
  char *p;

  copy_curr_dev();
  p = (char *) value;
  if (*p == '|')
    curr_dev.type = 2;
  if (*p == '>')
    curr_dev.type = 1;
  if ((*p == '|') || (*p == '>')) {
    p++;
    while (*p && isspace(*p))
      p++;
  }
  s = malloc(strlen(p) + 1);
  if (!s) {
    wlog(L_ERROR, "Malloc failed at allocating memory for device name");
    return 0;
  }
  strcpy(s, p);
  curr_dev.device = s;
  curr_dev.name = strdup(s);
  if (!curr_dev.name) {
    wlog(L_ERROR, "Malloc failed at allocating memory for device name");
    return 0;
  }
  return 1;
} /* add_device */

/* (re)read devices config */
int reread_devices_conf() {
  int res;
  tap_device *dev;

  destroy_all_devices();

  /* Initialize defaults */
  curr_dev.device = NULL;
  copy_curr_dev();

  res = read_config(file_dev_conf, devices_conf, DEVICE_CONF_NUM, section);
  if (!res) {
    wlog(L_ERROR, "Error in devices config: %s", config_error);
    return 0;
  }
  copy_curr_dev();
  if (curr_dev.init1)
    free(curr_dev.init1);
  if (curr_dev.init2)
    free(curr_dev.init2);
  if (curr_dev.dial_p)
    free(curr_dev.dial_p);
  if (curr_dev.dial_s)
    free(curr_dev.dial_s);
  if (curr_dev.passwd)
    free(curr_dev.passwd);
  wlog(L_DEBUG, "Devices config: %d lines", res);
  if (!head.stqh_first) {
    wlog(L_ERROR, "No devices found in config");
    return 0;
  }
  flag_fatal_error = 0;
  for (dev = head.stqh_first; dev; dev = dev->devp.stqe_next)
    if (!dev->master) {
      reconnect(dev);
      if (flag_fatal_error)
	return 0;
      if (dev->type == 0) {
	dev->status |= T_INIT1 | T_CRWAIT;
	if (dev->init1 && dev->init1[0])
	  queue_ack(dev, dev->init1);
	dev->tblock = 1;
      }
    }
  return 1;
} /* reread_devices_conf */

#define MAIN_CONF_NUM 18
conf_rec main_conf[MAIN_CONF_NUM] = {
  {PARAM("pidfile"), STRING, (void *) &file_pid,
   0, (void *) DEFAULT_PID_FILE, 1, std_STRING},
  {PARAM("lockdir"), STRING, (void *) &lock_dir,
   0, (void *) DEFAULT_LOCK_DIR, 1, std_STRING},
  {PARAM("maxqueuesize"), DEC_DIGITS, (void *) &maxpacks,
   0, (void *) 128, 1, std_DIGITS},
  {PARAM("maxqueuedelayed"), DEC_DIGITS, (void *) &max_wopacks,
   0, (void *) 128, 1, std_DIGITS},
  {PARAM("tnppaddress"), HEX_DIGITS, (void *) &my_tnpp_addr,
   0, (void *) 1, 1, std_DIGITS},
  {PARAM("devicefile"), STRING, (void *) &file_dev_conf,
   0, (void *) DEFAULT_DEV_CONFIG, 1, std_STRING},
  {PARAM("badmessages"), STRING, (void *) &file_bad,
   0, NULL, 1, std_STRING},
  {PARAM("sqlserver"), STRING, (void *) &sql_serv,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlport"), STRING, (void *) &sql_port,
   0, (void *) "5432", 1, std_STRING},
  {PARAM("sqldatabase"), STRING, (void *) &sql_base,
   0, (void *) "rtnppd", 1, std_STRING},
  {PARAM("sqluser"), STRING, (void *) &sql_user,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlpassword"), STRING, (void *) &sql_passwd,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("routingtable"), STRING, (void *) &file_route_table,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("extsocketin"), STRING, (void *) &u_sock_name,
   0, (void *) DEFAULT_SOCKET_IN_FILE, 1, std_STRING},
  {PARAM("extsocketout"), STRING, (void *) &u_sock_name_out,
   0, (void *) DEFAULT_SOCKET_OUT_FILE, 1, std_STRING},
  {PARAM("allowsendall"), BOOL_PAR, (void *) &allowsend,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("shedulerfunc"), HEX_DIGITS, (void *) &sheduler_func,
   0, (void *) 0x20, 1, std_DIGITS},
  {PARAM("runcommand"), STRING, (void *) &command_filter,
   0, (void *) NULL, 1, std_STRING}
};

/* (re)read fixed routing table from file */
int reread_route_table() {
  fixed_route *route;
  FILE *f;
  char buf[512];
  int res;
  char *p;
  char *s;

  while(routes.stqh_first) {
    route = routes.stqh_first;
    STAILQ_REMOVE_HEAD(&routes, next);
    free(route);
  }
  STAILQ_INIT(&routes);

  f = fopen(file_route_table, "r");
  if (!f) {
    wlog(L_ERROR, "Error %s (%d) at open routing table file",
	 strerror(errno), errno);
    return 0;
  }
  while (fgets(buf, sizeof(buf), f) != NULL) {
    if ((buf[0] == '\n') || !buf[0] || (buf[0] == ';') || (buf[0] == '#'))
      continue;
    route = (fixed_route *) malloc(sizeof(fixed_route));
    if (!route) {
      wlog(L_ERROR, "Malloc failed at allocating memory for route");
      return 0;
    }
    for (p = buf; (*p != ' ') && (*p != '\t') && *p; p++);
    *p = '\0';
    route->from = strtol(buf, NULL, 16);
    for (s = p + 1; *s && ((*s == ' ') || (*s == '\t')); s++);
    for (p = s; (*p != ' ') && (*p != '\t') && *p; p++);
    if (!(*p)) {
      wlog(L_ERROR, "Bad line format in routing table file");
      return 0;
    }
    *p = '\0';
    strncpy(route->to, s, sizeof(route->to));
    route->to[sizeof(route->to) - 1] = '\0';
    for (s = p + 1; *s && ((*s == ' ') || (*s == '\t')); s++);
    for (p = s + strlen(s); *p < ' '; *p-- = '\0');
    strncpy(route->destination, s, sizeof(route->destination));
    route->destination[sizeof(route->destination) - 1] = '\0';
    STAILQ_INSERT_TAIL(&routes, route, next);
  }
  fclose(f);
  return 1;
} /* reread_route_table */

/* (re)read main config */
int reread_main_conf() {
  int res;
  FILE *f;

  if (file_pid) {
    unlink(file_pid);
    free(file_pid);
  }
  if (lock_dir)
    free(lock_dir);

  if (sql_serv)
    free(sql_serv);
  if (sql_user)
    free(sql_user);
  if (sql_passwd)
    free(sql_passwd);
  if (sql_base)
    free(sql_base);
  if (sql_port)
    free(sql_port);
  if (u_sock_name) {
    unlink(u_sock_name);
    free(u_sock_name);
  }
  if (u_sock_name_out)
    free(u_sock_name_out);
  if (file_dev_conf)
    free(file_dev_conf);
  if (file_route_table)
    free(file_route_table);

  res = read_config(file_main_conf, main_conf, MAIN_CONF_NUM, section);
  if (!res) {
    wlog(L_ERROR, "Error in main config: %s", config_error);
    return 0;
  }
  wlog(L_DEBUG, "Main config: %d lines", res);

  /* Bind unix socket */
  memset(&u_sun, 0, sizeof(u_sun));
  u_sun.sun_family = AF_UNIX;
  strncpy(u_sun.sun_path, u_sock_name, sizeof(u_sun.sun_path));
  tnpp_in = socket(AF_UNIX, SOCK_DGRAM, 0);
  tnpp_out = socket(AF_UNIX, SOCK_DGRAM, 0);
  if ((tnpp_in < 0) || (tnpp_out < 0)) {
    wlog(L_ERROR, "Error %s (%d) at create unix socket",
	 strerror(errno), errno);
    return 0;
  }
  unlink(u_sock_name);
  res = bind(tnpp_in, (struct sockaddr *) &u_sun, strlen(u_sun.sun_path) + 2);
  if (res < 0) {
    wlog(L_ERROR, "Error %s (%d) at bind unix socket",
	 strerror(errno), errno);
    return 0;
  }
  if (allowsend)
    chmod(u_sock_name, 0622);
  else
    chmod(u_sock_name, 0600);
  memset(&u_sun, 0, sizeof(u_sun));
  u_sun.sun_family = AF_UNIX;
  strncpy(u_sun.sun_path, u_sock_name_out, sizeof(u_sun.sun_path));

  if (!reread_devices_conf())
    return 0;

  unlink(file_pid);
  f = fopen(file_pid, "w");
  if (!f)
    wlog(L_WARNING, "Error %s (%d) at create PID-file",
	 strerror(errno), errno);
  else {
    fprintf(f, "%d\n", getpid());
    fclose(f);
  }
  if (file_route_table) {
    use_file_table = 1;
    if (!reread_route_table())
      return 0;
  } else
    if (to_sql_connect())
      return 0;
  return 1;
} /* reread_main_conf */

void to_sql_disconnect() {
#ifdef USE_MSQL
  if (use_sql)
    msqlClose(sock_sql);
#else
#ifdef USE_POSTGRESQL
  if (use_sql)
    PQfinish(conn_sql);
#else
#ifdef USE_MYSQL
  if (use_sql) {
    mysql_close(conn_sql);
    conn_sql = NULL;
  }
#endif
#endif
#endif
  use_sql = 0;
} /* to_sql_disconnect */

/* Close sql-connection, sockets etc */
void close_fds() {
  if (tnpp_in >= 0) {
    unlink(u_sock_name);
    close(tnpp_in);
  }
  if (tnpp_out >= 0)
    close(tnpp_out);
  to_sql_disconnect();
} /* close_fds */

void usage() {
  char thishostname[512];

  gethostname(thishostname, 512);
  printf("\nrtapd v1.7 by Vasim V.\n");
  printf("\nUsage: rtapd [-ds] [-C CONFIG] [-S SECTION] [-x LOGLEVEL]\n");
  printf("\nOptions:\n");
  printf("  -d\t\tDo not daemonize\n");
  printf("  -s\t\tUse stderr instead syslog\n");
  printf("  -C CONFIG\tUse config file CONFIG instead '%s'\n",
	 DEFAULT_MAIN_CONFIG);
  printf("  -S SECTION\tUse section [SECTION] in config files instead [%s]\n",
	 thishostname);
  printf("  -p FILEDUMP\tDump all configuration data in file FILEDUMP\n");
  printf("  -x LOGLEVEL\tLog at level LOGLEVEL (default is 2):\n");
  printf("\t\t\t0 - No logging\n");
  printf("\t\t\t1 - Log only errors\n");
  printf("\t\t\t2 - Log errors and warnings\n");
  printf("\t\t\t3 - Log some info about packets, etc\n");
  printf("\t\t\t4 - Log all info\n");
  printf("\t\t\t5 - Log all info and all data\n\n");
} /* usage */

/* main */
int main(int argc, char *argv[]) {
  extern char *optarg;
  extern int optind;
  int ch;
  int levels[6] = {L_FATAL, L_ERROR, L_WARNING, L_INFO, L_DEBUG, L_DEBUG};
  int level;
  int res;
  char *filedump;

  nodaemon = 0;
  log_stderr = 0;
  file_main_conf = DEFAULT_MAIN_CONFIG;
  loglevel = L_WARNING;
  section = NULL;
  log_all_data = 0;
  filedump = NULL;

  /* Parse command-line options */
  while ((ch = getopt(argc, argv, "dsC:x:S:p:")) != -1)
    switch (ch) {
    case 'd':
      nodaemon = 1;
      break;
    case 's':
      log_stderr = 1;
      break;
    case 'C':
      file_main_conf = optarg;
      break;
    case 'p':
      filedump = optarg;
      break;
    case 'x':
      level = atoi(optarg);
      if ((level >= 0) && (level <= 5))
	loglevel = levels[level];
      else {
	usage();
	exit(1);
      }
      if (level == 5)
	log_all_data = 1;
      break;
    case 'S':
      section = optarg;
      break;
    case '?':
      usage();
      exit(1);
    }
  if (!log_stderr)
    openlog("rtapd", LOG_PID | LOG_NDELAY, LOG_USER);

  file_dev_conf = NULL;
  file_route_table = NULL;
  defaultroute = NULL;
  flag_default = 0;
  lock_dir = NULL;
  file_pid = NULL;
  sql_serv = NULL;
  sql_base = NULL;
  sql_port = NULL;
  sql_user = NULL;
  sql_passwd = NULL;
  u_sock_name = NULL;
  u_sock_name_out = NULL;
  tnpp_in = -1;
  tnpp_out = -1;
  STAILQ_INIT(&head);
  STAILQ_INIT(&wpacks);
  nwpacks = 0;
  STAILQ_INIT(&routes);
  use_sql = 0;

  if (!nodaemon) {
    res = fork();
    if (res > 0)
      exit(0);
    setsid();
  }

  do {
    if (!reread_main_conf())
      exit(2);
    if (filedump)
      dump_config(filedump);
    loop();
    close_fds();
  } while (!flag_fatal_error);
  destroy_all_devices();
  if (!log_stderr)
    closelog();
  exit(1);
} /* main */
