/* Definitions for rtapd */
/* $Id: rtapd.h,v 1.11 2001/12/02 06:58:46 vasim Exp $ */

#include "queue.h"

#define MAX_PACKET_SIZE 4160
#define DEFAULT_MAIN_CONFIG "/usr/local/etc/rtnppd/rtapd.conf"
#define DEFAULT_DEV_CONFIG "/usr/local/etc/rtnppd/tap_dev.conf"
#define DEFAULT_PID_FILE "/var/run/rtapd.pid"
#define DEFAULT_SOCKET_IN_FILE "/var/run/rtnppd.queue.out"
#define DEFAULT_SOCKET_OUT_FILE "/var/run/rtnppd.queue.in"
#define DEFAULT_LOCK_DIR "/var/spool/lock"
#define SERIAL_TABLE_NUM 64

#define SOH '\01'
#define STX '\02'
#define ETX '\03'
#define EOT '\04'
#define ENQ '\05'
#define ACK '\06'
#define NAK '\025'
#define ETB '\027'
#define CAN '\030'
#define RS '\036'
#define US '\037'
#define SUB '\032'
#define CR '\015'
#define LF '\012'
#define ESC '\033'

#define P_MESSAGE 0x000001    /* This packet is dialog message (ack,
			       * greetings, etc) */
#define P_DATA 0x000002       /* This packet is data */
#define P_SENDING 0x000004    /* packet are sending */
#define P_WAITACK 0x000008    /* Waiting for ACK */

typedef struct tap_packet {
  int status;     /* Status of this packet (one packet - one block( */
  int offset;     /* Offset sending data in packet */
  int size;       /* Size of this packet */
  unsigned char data[270];  /* Data of this packet */
  int nc;         /* Number of retries of calls to this packet */
  unsigned char *msg;         /* Pointer to original message */
  int msglen;
  STAILQ_ENTRY(tap_packet) packp;
} tap_packet;

#define T_DIALSTATE 0x000001  /* Waiting for connect */
#define T_CONNECTED 0x000002  /* Connected */
#define T_CRWAIT 0x000004     /* Waiting for end-of-line (CR) */
#define T_INIT1 0x000008      /* Initializing (first stage) */
#define T_INIT2 0x000010      /* Initializing (second stage) */
#define T_IDENTWAIT 0x000020  /* Waiting for identify string */
#define T_BLOCKWAIT 0x000040  /* Waiting for block */
#define T_BLOCKEND 0x000080   /* Waiting for block's end */
#define T_ACKWAIT 0x000100    /* Waiting for acknowledgement */
#define T_CALLEND 0x000200    /* Waiting for call's end */
#define T_CALLOUT 0x000400    /* We're sending, not receiving */
#define T_BUSY 0x000800       /* Port for this device is busy */

typedef struct tap_device {
  int fd;         /* File descriptor */
  char *device;   /* Device filename */
  char *name;     /* Name of this device (for logs and route tables) */
  int status;     /* Status of this device */
  int type;       /* 0 - serial devices */
  char *init1;    /* Initialize string 1 */
  char *init2;    /* Initialize string 2 */
  char *dial_p;   /* Dial prefix */
  char *dial_s;   /* Dial suffix */
  char phone[32]; /* Phone number */
  char *passwd;   /* Callout password for this device */
  int next_multi; /* Number of next unused device in '*' list */
  int clocal;     /* Don't check CD */
  int dial;       /* Send dail_p + phone + dial_s before connect */
  int parity;     /* Use even parity */
  int autodetect; /* Autodetect parity */
  int baud;       /* Serial speed */
  int xon_xoff;   /* Use XON/XOFF data flow control */
  int crtscts;    /* Use RTS/CTS flow control */
  int dtrtoggle;  /* Toggle DTR */
  int lfcr;       /* Convert CR->LF+CR */
  int crlf;       /* Convert LF->CR+LF */
  int max_blocks; /* Maximum number of blocks for this device */
  int max_trans;  /* Maximum number of transactions */
  int max_r_len;  /* Maximum length of message (runtime) */
  int n_trans;    /* Number of current transaction */
  int max_r_trans;/* Maximum number of transactions (runtime) */
  unsigned char buf[MAX_PACKET_SIZE];  /* Receive buffer */
  int offset;     /* Offset in receive buffer */
  int pos_block;  /* Offset of receiving block */
  int pos_line;   /* Begin of last line */
  STAILQ_HEAD(packhead, tap_packet) head;
  int npacks;     /* Number of blocks in queue */
  struct tap_device *master;        /* Pointer to master device */
  STAILQ_ENTRY(tap_device) devp;
  int t1, t2, t3, t4, t5;           /* Timers (see TAP 1.8 specs) */
  int tr;                           /* Timeout on receive */
  int tblock;                       /* Timeout on device error */
  int tsess;                        /* Timeout on session */
  int i_t1, i_t2, i_t3, i_t4, i_t5; /* Default values of timers */
  int i_tr;
  int i_tblock;
  int i_tsess;
  int n1, n2, n3;                   /* Counters (see TAP 1.8 specs) */
  int nr;                           /* Numbers of retries */
  int i_n1, i_n2, i_n3;             /* Default values of counters */
  int i_nr, i_nc;
  int lastack;                      /* Last ACK (0) or NAK (1) or RS (2) */
  int lastdial;			    /* 1 if last dial was unsuccessful */
  int no_incoming;		    /* 1 if the rtapd must not answer */
} tap_device;
