/* vsnppd by Vasim V. */
/* -Id: tnpppage.c,v 1.6 1999/04/28 09:41:10 vasim Exp - */
/* $Id: vsnppd.c,v 1.3 1999/09/23 12:51:10 vasim Exp $ */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include "crc.h"
#include "readcfg.h"

#define MAX_PACKET_SIZE 4160
#define DEFAULT_CONFIG_NAME "/usr/local/etc/rtnppd/rtapd.conf"
#define DEFAULT_SOCK_NAME "/var/run/rtnppd.queue.out"

int my_tnpp_addr;

char page_func = '@';


int tnpp_out;

int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		     int len, int tnpp_to, int tnpp_from);

/* SNPP Client Commands */

#define PAGER        "PAGE"
#define MESSAGE      "MESS"
#define SEND         "SEND"
#define QUIT         "QUIT"
#define RESET        "RESE"
#define HELP         "HELP"

/* Responses from SNPP server to client */

#define SNPP_OK      "250 OK"
#define SNPP_RESET   "250 Reset OK"
#define SNPP_SENT    "250 Page Sent"
#define SNPP_BADPIN  "550 Failed,"
#define SNPP_NOTSENT "554 Failed,"
#define SNPP_ENTERR  "503 Error, Already Entered"
#define SNPP_ERRINC  "503 Error, Incomplete Info"
#define SNPP_OKCLOS  "221 OK, Goodbye"
#define SNPP_TIMEOUT "421 Timeout, Goodbye"
#define SNPP_ERRCLOS "421 ERROR, Goodbye"
#define SNPP_DOWN    "421 SERVER DOWN, Goodbye"


/* Send TNPP packet via unix socket */
int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		      int len, int tnpp_to, int tnpp_from) {
  unsigned char buf[MAX_PACKET_SIZE];
  int length, crc;

  if (len > (MAX_PACKET_SIZE - 14))
    return;
  snprintf(buf, sizeof(buf) - 1, "%c%04X%02X%04X%02X%c", '\01', tnpp_to, 20,
	   tnpp_from, 0, '\02');
  length = strlen(buf);
  memcpy(buf + strlen(buf), data, len);
  length = length + len;
  buf[length] = '\03';
  length++;
  crc = crc16(buf, length);
  buf[length] = (unsigned char) (crc & 0xff);
  buf[length + 1] = (unsigned char) ((crc >> 8) & 0xff);
  length = length + 2;

  /* Send packet now */
  return (sendto(fd, buf, length, 0, (struct sockaddr *) sun,
		 strlen(sun->sun_path) + 2));
} /* send_tnpp_packet */  

void usage() {
  printf("\nvsnppd v1.1 by Vasim V.\n");
  printf("\nUsage:\tvsnppd [-t TNPP-ADDR -s SOCKET] [-c FUNC] [-C CONFIG]\n");
  printf("  -t TNPP-ADDR\tSend message to device with address TNPP-ADDR (default is from %s)\n",
	DEFAULT_CONFIG_NAME);
  printf("  -s SOCKET\tUse unix socket SOCKET (default is from %s)\n",
	 DEFAULT_CONFIG_NAME);
  printf("  -c FUNC\tSend message with function code FUNC (default is 0x40)\n");
  printf("  -C CONFIG\tRead config file CONFIG instead %s\n\n",
	DEFAULT_CONFIG_NAME);
} /* usage */

char *u_sock_name_out = DEFAULT_SOCK_NAME;

/* Unused variables */
char *file_pid = NULL;
char *lock_dir = NULL;
char *u_sock_name = NULL;
char *file_dev_conf = NULL;
char *file_bad = NULL;
char *sql_serv = NULL;
char *sql_port = NULL;
char *sql_base = NULL;
char *sql_user = NULL;
char *sql_passwd = NULL;
char *file_route_table = NULL;
char *command_filter = NULL;
int sheduler_func;
int maxpacks;
int max_wopacks;
int allowsend;

#define MAIN_CONF_NUM 18
conf_rec main_conf[MAIN_CONF_NUM] = {
  {PARAM("pidfile"), STRING, (void *) &file_pid,
   0, (void *) "x", 1, std_STRING},
  {PARAM("lockdir"), STRING, (void *) &lock_dir,
   0, (void *) "x", 1, std_STRING},
  {PARAM("maxqueuesize"), DEC_DIGITS, (void *) &maxpacks,
   0, (void *) 128, 1, std_DIGITS},
  {PARAM("maxqueuedelayed"), DEC_DIGITS, (void *) &max_wopacks,
   0, (void *) 128, 1, std_DIGITS},
  {PARAM("tnppaddress"), HEX_DIGITS, (void *) &my_tnpp_addr,
   0, (void *) 1, 1, std_DIGITS},
  {PARAM("devicefile"), STRING, (void *) &file_dev_conf,
   0, (void *) "x", 1, std_STRING},
  {PARAM("badmessages"), STRING, (void *) &file_bad,
   0, NULL, 1, std_STRING},
  {PARAM("sqlserver"), STRING, (void *) &sql_serv,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlport"), STRING, (void *) &sql_port,
   0, (void *) "5432", 1, std_STRING},
  {PARAM("sqldatabase"), STRING, (void *) &sql_base,
   0, (void *) "rtnppd", 1, std_STRING},
  {PARAM("sqluser"), STRING, (void *) &sql_user,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlpassword"), STRING, (void *) &sql_passwd,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("routingtable"), STRING, (void *) &file_route_table,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("extsocketin"), STRING, (void *) &u_sock_name_out,
   0, (void *) DEFAULT_SOCK_NAME, 1, std_STRING},
  {PARAM("extsocketout"), STRING, (void *) &u_sock_name,
   0, (void *) "x", 1, std_STRING},
  {PARAM("allowsendall"), BOOL_PAR, (void *) &allowsend,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("shedulerfunc"), HEX_DIGITS, (void *) &sheduler_func,
   0, (void *) 0x20, 1, std_DIGITS},
  {PARAM("runcommand"), STRING, (void *) &command_filter,
   0, (void *) NULL, 1, std_STRING}
};

/* main */
int main(int argc, char *argv[]) {
  extern char *optarg;
  extern int optind;
  char id[11];
  int dest = 0;
  int ch;
  unsigned char pbuf[MAX_PACKET_SIZE];
  unsigned char buf[4076];
  unsigned char *p;
  unsigned char *pc;
  char sbf[5];
  struct sockaddr_un u_sun;
  unsigned char c;
  int offset;
  int i;
  int res;
  int len;
  char bufid[13];
  char bufl[1024];
  char *configfile;
  int config;
  int status;

  config = 0;
  configfile = DEFAULT_CONFIG_NAME;
  id[0] = '\0';
  /* Parse command-line options */
  while ((ch = getopt(argc, argv, "t:s:c:C:")) != -1)
    switch (ch) {
    case 't':
      offset = 0;
      if ((*optarg == '0') && (strlen(optarg) > 1)
	  && (toupper(*(optarg + 1)) == 'X'))
	offset = 2;
      if (toupper(*optarg) == 'X')
	offset = 1;
      for (i = 0; (i < 4) && optarg[i + offset]; i++)
	sbf[i] = toupper(optarg[i + offset]);
      sbf[i] = '\0';
      dest = strtol(sbf, NULL, 16);
      config = 1;
      break;
    case 's':
      u_sock_name_out = optarg;
      break;
    case 'c':
      offset = 0;
      if ((*optarg == '0') && (strlen(optarg) > 1)
	  && (toupper(*(optarg + 1)) == 'X'))
	offset = 2;
      if (toupper(*optarg) == 'X')
	offset = 1;
      for (i = 0; (i < 2) && optarg[i + offset]; i++)
	sbf[i] = toupper(optarg[i + offset]);
      sbf[i] = '\0';
      page_func = strtol(sbf, NULL, 16);
      break;
    case 'C':
      configfile = optarg;
      config = 0;
      break;
    case '?':
      usage();
      exit(1);
    }

  if (!config) {
    u_sock_name_out = NULL;
    res = read_config(configfile, main_conf, MAIN_CONF_NUM, NULL);
    if (!res) {
      fprintf(stderr, "Error in config '%s': %s\n", configfile, config_error);
      exit(1);
    }
    dest = my_tnpp_addr;
  } else
    my_tnpp_addr = dest;

  memset(&u_sun, 0, sizeof(u_sun));
  u_sun.sun_family = AF_UNIX;
  strncpy(u_sun.sun_path, u_sock_name_out, sizeof(u_sun.sun_path));
  tnpp_out = socket(AF_UNIX, SOCK_DGRAM, 0);
  if (tnpp_out < 0) {
    fprintf(stderr, "Error %s (%d) at creating unix socket\n",
	    strerror(errno), errno);
    exit(2);
  }

  printf("220 SNPP Gateway Ready\r\n");
  status = 0;
  id[0] = '\0';
  buf[0] = '\0';
  fflush(stdout);
  while (fgets(bufl, sizeof(bufl), stdin)) {
    if (status) {
      if ((bufl[0] == '.') && ((bufl[1] == '\r') || (bufl[1] == '\n'))) {
	status = 0;
	printf("250 Message OK\r\n");
      } else {
	len = strlen(buf);
	if (sizeof(buf) > (strlen(bufl) + len + 1))
	  strcpy(buf + len, bufl);
      }
    } else {
      bufl[0] = toupper(bufl[0]);
      bufl[1] = toupper(bufl[1]);
      bufl[2] = toupper(bufl[2]);
      bufl[3] = toupper(bufl[3]);
      if (!strncmp(bufl, HELP, 4)) {
	printf("214 vsnppd v1.1 gateway for rtapd by Vasim V.\r\n");
	printf("214 SNPP commands:\r\n");
	printf("214 PAGE, MESS, SEND, QUIT, RESE, HELP\r\n");
	printf("214 SNPP version 1 (with level 2 minimum extensions).\r\n");
	printf("214 See RFC-1861 for more details\r\n");
	printf("%s\r\n", SNPP_OK);
      } else if (!strncmp(bufl, RESET, 4)) {
	id[0] = '\0';
	buf[0] = '\0';
	printf("%s\r\n", SNPP_RESET);
      } else if (!strncmp(bufl, QUIT, 4)) {
	printf("%s\r\n", SNPP_OKCLOS);
	exit(0);
      } else if (!strncmp(bufl, PAGER, 4)) {
	if (id[0]) {
	  printf("%s\r\n", SNPP_ENTERR);
	} else {
	  p = strchr(bufl, ' ');
	  if (!p)
	    printf("%s\r\n", SNPP_ERRINC);
	  else {
	    for (pc = p; (*pc == ' '); pc++);
	    if (*pc <= ' ')
	      printf("%s\r\n", SNPP_ERRINC);
	    else {
	      for (p = pc + 1; (*p > ' '); p++);
	      if (*p != '\0')
		*p = '\0';
	      strncpy(id, pc, sizeof(id));
	      id[10] = '\0';
	      printf("250 Pager ID Accepted\r\n");
	    }
	  }
	}
      } else if (!strncmp(bufl, MESSAGE, 4)) {
	if (buf[0])
	  printf("%s\r\n", SNPP_ENTERR);
	else {
	  p = strchr(bufl, ' ');
	  if (!p)
	    printf("%s\r\n", SNPP_ERRINC);
	  else {
	    for (pc = p; (*pc == ' '); pc++);
	    if (*pc <= ' ')
	      printf("%s\r\n", SNPP_ERRINC);
	    else {
	      for (p = pc + 1; (*p >= ' '); p++);
	      if (*p != '\0')
		*p = '\0';
	      strncpy(buf, pc, sizeof(buf));
	      buf[sizeof(buf) - 1] = '\0';
	      printf("%s\r\n", SNPP_OK);
	    }
	  }
	}
      } else if (!strncmp(bufl, SEND, 4)) {
	if (!id[0] || !buf[0])
	  printf("%s\r\n", SNPP_ERRINC);
	else {
	  p = buf;
	  pc = pbuf + 12;
	  for (i = 0; (i < 4078) && (pc < (pbuf + sizeof(pbuf) - 2)); i++) {
	    c = p[i];
	    if (!c)
	      break;
	    if ((c != '\01') && (c != '\02') && (c != '\03') && (c != '\04')
		&& (c != '\05') && (c != '\06') && (c != '\025')
		&& (c != '\027') && (c != '\030') && (c != '\036')
		&& (c != '\037') && (c != '\032') && (c != '\015')
		&& (c != '\012') && (c != '\033'))
	      *pc++ = c;
	    else {
	      *pc++ = '\032';
	      *pc++ = c + 0x40;
	    }
	  }
	  len = pc - pbuf - 12;
	  sprintf(bufid, "B%c%- 10s", page_func, id);
	  memcpy(pbuf, bufid, 12);
	  *pc = '\027';
	  if (send_tnpp_packet(tnpp_out, &u_sun, pbuf, len + 12,
			       dest, 0) <= 0)
	    printf("%s system error '%s' (%d) at send\r\n", SNPP_NOTSENT,
		   strerror(errno), errno, u_sock_name_out);
	  else
	    printf("%s\r\n", SNPP_SENT);
	  id[0] = '\0';
	  buf[0] = '\0';
	}
      } else if (!strncmp(bufl, "LOGI", 4)) {
	/* Nothing to do with login request */
	printf("250 Login Accepted\r\n");
      } else if (!strncmp(bufl, "DATA", 4)) {
	if (buf[0])
	  printf("%s\r\n", SNPP_ERRINC);
	else {
	  status = 1;
	  printf("354 Begin Input; End with <CRLF>'.'<CRLF>\r\n");
	}
      } else if (!strncmp(bufl, "LEVE", 4)) {
	/* Nothing to do with LEVEL command */
	printf("250 OK, Alternate Service Level Accepted\r\n");
      } else if (!strncmp(bufl, "ALER", 4)
		 || !strncmp(bufl, "COVE", 4) || !strncmp(bufl, "HOLD", 4)
		 || !strncmp(bufl, "CALL", 4) || !strncmp(bufl, "SUBJ", 4)
		 || !strncmp(bufl, "2WAY", 4) || !strncmp(bufl, "PING", 4)
		 || !strncmp(bufl, "EXPT", 4) || !strncmp(bufl, "NOQU", 4)
		 || !strncmp(bufl, "ACKR", 4) || !strncmp(bufl, "RTYP", 4)
		 || !strncmp(bufl, "MCRE", 4) || !strncmp(bufl, "MSTA", 4)
		 || !strncmp(bufl, "KTAG", 4)) {
	printf("500 Command Not Implemented, Try Again\r\n");      
      } else{
	printf("500 Command Not Implemented, Try Again\r\n");
      }
    }
    fflush(stdout);
  }

  close(tnpp_out);
  exit(0);
} /* main */
